from __future__ import print_function, division, absolute_import
import os
import pickle

import numpy as np

from odin.fuel import MmapData
from odin.ml.gmm_tmat import GMM, Tmatrix, _split_jobs
from odin.ml.base import BaseEstimator, TransformerMixin, DensityMixin
from odin.utils import (mpi, batching, Progbar, crypto, is_primitives, ctext,
                        uuid, UnitTimer)

# ===========================================================================
# Helper
# ===========================================================================
def _extract_zero_and_first_stats(X, sad, indices, gmm, z_path, f_path, name_path):
  n_samples = X.shape[0]
  # indices is None, every row is single sample (utterance or image ...)
  if indices is None:
    if os.path.exists(z_path):
      os.remove(z_path)
    if os.path.exists(f_path):
      os.remove(f_path)
    Z = MmapData(path=z_path, dtype='float32',
                 shape=(n_samples, gmm.nmix), read_only=False)
    F = MmapData(path=f_path, dtype='float32',
                 shape=(n_samples, gmm.feat_dim * gmm.nmix),
                 read_only=False)
    jobs, _ = _split_jobs(n_samples, ncpu=mpi.cpu_count(),
                       device='cpu', gpu_factor=1)

    def map_transform(start_end):
      start, end = start_end
      for i in range(start, end):
        # removed by SAD
        if sad is not None and not bool(sad[i]):
          yield None, None, None
        else:
          z, f = gmm.transform(X[i][np.newaxis, :],
                               zero=True, first=True, device='cpu')
          yield i, z, f
    prog = Progbar(target=n_samples,
                   print_report=True, print_summary=False,
                   name="Extracting zero and first order statistics")
    for i, z, f in mpi.MPI(jobs, map_transform,
                           ncpu=None, batch=1):
      if i is not None: # i None means removed by SAD
        Z[i] = z
        F[i] = f
      prog.add(1)
    Z.flush(); Z.close()
    F.flush(); F.close()
  # use directly the transform_to_disk function
  else:
    gmm.transform_to_disk(X, indices=indices, sad=sad,
                          pathZ=z_path,
                          pathF=f_path,
                          name_path=name_path,
                          dtype='float32', device=None, ncpu=None,
                          override=True)

# ===========================================================================
# Fast combined GMM-Tmatrix training for I-vector extraction
# ===========================================================================
class Ivector(DensityMixin, BaseEstimator, TransformerMixin):
  """ Ivector extraction using GMM and T-matrix """

  def __init__(self, path, nmix=None, tv_dim=None,
               nmix_start=1, niter_gmm=16, niter_tmat=16,
               allow_rollback=True, exit_on_error=False,
               downsample=1, stochastic_downsample=True,
               device='gpu', ncpu=1, gpu_factor_gmm=80, gpu_factor_tmat=3,
               dtype='float32', seed=5218, name=None):
    super(Ivector, self).__init__()
    # ====== auto store arguments ====== #
    for key, val in locals().items():
      if key in ('self', 'path', 'seed'):
        continue
      setattr(self, key, val)
    # ====== create random generator ====== #
    self._rand = np.random.RandomState(seed=seed)
    # ====== check path ====== #
    path = str(path)
    if not os.path.exists(path):
      os.mkdir(path)
    elif not os.path.isdir(path):
      raise ValueError("Path to '%s' is not a directory" % str(path))
    self._path = path
    self._gmm = None
    self._tmat = None

  # ==================== properties ==================== #
  @property
  def gmm(self):
    if self._gmm is None:
      if os.path.exists(self.gmm_path):
        with open(self.gmm_path, 'rb') as f:
          self._gmm = pickle.load(f)
        assert self._gmm.nmix == self.nmix, \
        "Require GMM with %d components, but found %s, at path: '%s'" % \
        (self.nmix, str(self._gmm), self.gmm_path)
      else:
        self._gmm = GMM(nmix=self.nmix, niter=self.niter_gmm, dtype=self.dtype,
                        downsample=self.downsample,
                        stochastic_downsample=self.stochastic_downsample,
                        device=self.device,
                        ncpu=self.ncpu, gpu_factor=self.gpu_factor_gmm,
                        seed=5218,
                        path=self.gmm_path,
                        name="IvecGMM_%s" % (self.name if self.name is not None else
                                             str(self._rand.randint(10e8))))
    return self._gmm

  @property
  def tmat(self):
    if self._tmat is None:
      if os.path.exists(self.tmat_path):
        with open(self.tmat_path, 'rb') as f:
          self._tmat = pickle.load(f)
        assert self._tmat.tv_dim == self.tv_dim, \
        "Require T-matrix with %d dimensions, but found %s, at path: '%s'" % \
        (self.tv_dim, str(self._tmat), self.tmat_path)
      else:
        self._tmat = Tmatrix(tv_dim=self.tv_dim, gmm=self.gmm,
                             niter=self.niter_tmat, dtype=self.dtype,
                             device=self.device,
                             ncpu=self.ncpu, gpu_factor=self.gpu_factor_tmat,
                             cache_path='/tmp', seed=5218,
                             path=self.tmat_path,
                             name='IvecTmat_%s' % (self.name if self.name is not None else
                                                   str(self._rand.randint(10e8))))
    return self._tmat

  @property
  def path(self):
    return self._path

  @property
  def gmm_path(self):
    return os.path.join(self.path, 'gmm.pkl')

  @property
  def tmat_path(self):
    return os.path.join(self.path, 'tmat.pkl')

  @property
  def z_path(self):
    """ Path to zero-th order statistics of the training data"""
    return os.path.join(self.path, 'zstat_train')

  @property
  def f_path(self):
    """ Path to first order statistics of the training data """
    return os.path.join(self.path, 'fstat_train')

  @property
  def ivec_path(self):
    """ Path to first order statistics of the training data """
    return os.path.join(self.path, 'ivec_train')

  @property
  def name_path(self):
    """ In case indices is given during training, the order of
    processed files is store at this path """
    return os.path.join(self.path, 'name_train')

  @property
  def feat_dim(self):
    return self.gmm.feat_dim

  # ==================== state query ==================== #
  @property
  def is_gmm_fitted(self):
    return self.gmm.is_fitted

  @property
  def is_tmat_fitted(self):
    return self.is_gmm_fitted and self.tmat.is_fitted

  @property
  def is_fitted(self):
    return self.is_gmm_fitted and self.is_tmat_fitted

  # ====== getter ====== #
  def get_z_path(self, name=None):
    """ Return the path the zero-order statistics
    according to the given name as identification during
    `Ivector.transform`
    If name is None, return `Ivector.z_path`
    """
    if name is None:
      return self.z_path
    return os.path.join(self.path, 'zstat_%s' % name)

  def get_f_path(self, name):
    """ Return the path the first-order statistics
    according to the given name as identification during
    `Ivector.transform`
    If name is None, return `Ivector.f_path`
    """
    if name is None:
      return self.f_path
    return os.path.join(self.path, 'fstat_%s' % name)

  def get_i_path(self, name):
    """ Return the path the extracted i-vectors
    according to the given name as identification during
    `Ivector.transform`
    If name is None, return `Ivector.ivec_path`
    """
    if name is None:
      return self.ivec_path
    return os.path.join(self.path, 'ivec_%s' % name)

  def get_name_path(self, name):
    """ Return the path of the name list if indices is used
    according to the given name as identification during
    `Ivector.transform`
    If name is None, return `Ivector.name_path`
    """
    if name is None:
      return self.name_path
    return os.path.join(self.path, 'name_%s' % name)

  # ==================== sklearn methods ==================== #
  def fit(self, X, indices=None, sad=None,
          refit_gmm=False, refit_tmat=False,
          extract_ivecs=False, keep_stats=False):
    """
    Parameters
    ----------
    X : ndarray
      Training data [n_samples, n_features]

    indices : {Mapping, tuple, list}
      in case the data is given by a list of files, `indices`
      act as file indicator mapping from
      'file_name' -> (start_index_in_X, end_index_in_X)
      This mapping can be provided by a dictionary, or list of
      tuple.
      Note: the order provided in indices will be preserved

    sad : ndarray
      inspired by the "Speech Activity Detection" (SAD) indexing,
      this array is indicator of which samples will be taken into
      training; the shape should be [n_samples,] or [n_samples, 1]

    refit_gmm : bool
      if True, re-fit the GMM even though it is fitted,
      consequently, the T-matrix will be re-fitted

    refit_tmat : bool
      if True, re-fit the T-matrix even though it is fitted

    extract_ivecs : bool
      if True, extract the i-vector for training data

    keep_stats : bool
      if True, keep the zero and first order statistics.
      The first order statistics could consume huge amount
      of disk space. Otherwise, they are deleted after training
    """
    new_gmm = (not self.gmm.is_fitted or refit_gmm)
    # ====== clean error files ====== #
    if os.path.exists(self.z_path):
      Z = MmapData(self.z_path, read_only=True)
      if Z.shape[0] == 0: # empty file
        os.remove(self.z_path)
      Z.close()
    if os.path.exists(self.f_path):
      F = MmapData(self.f_path, read_only=True)
      if F.shape[0] == 0: # empty file
        os.remove(self.f_path)
      F.close()
    if os.path.exists(self.ivec_path):
      ivec = MmapData(self.ivec_path, read_only=True)
      if ivec.shape[0] == 0: # empty file
        os.remove(self.ivec_path)
      ivec.close()
    # ====== Training the GMM first ====== #
    if new_gmm:
      input_data = [X]
      if sad is not None:
        input_data.append(sad)
      if indices is not None:
        input_data.append(indices)
      self.gmm.fit(input_data)
    # ====== some fun, and confusing logics ====== #
    # GMM need to be fitted before creating T-matrix model
    new_tmat = (not self.tmat.is_fitted or new_gmm or refit_tmat)
    # New I-vector is need when:
    # - only when `extract_ivecs=True`
    # - and new T-matrix is trained but no I-vector is extracted
    new_ivec = extract_ivecs and \
    (new_tmat or not os.path.exists(self.ivec_path))
    # new stats is only needed when
    # - GMM is updated
    # - training new Tmatrix and the Z and F not exist
    # - extracting new I-vector and the Z and F not exist
    if not new_gmm and \
    (os.path.exists(self.z_path) and os.path.exists(self.f_path)):
      new_stats = False
    else:
      new_stats = new_gmm or new_tmat or new_ivec
    # ====== extract the statistics ====== #
    if new_stats:
      _extract_zero_and_first_stats(X=X, sad=sad, indices=indices, gmm=self.gmm,
                                    z_path=self.z_path, f_path=self.f_path,
                                    name_path=self.name_path)
    # ====== Training the T-matrix and extract i-vector ====== #
    if new_tmat or new_ivec:
      Z = MmapData(path=self.z_path, read_only=True)
      F = MmapData(path=self.f_path, read_only=True)
      if new_tmat:
        self.tmat.fit((Z, F))
      if new_ivec:
        self.tmat.transform_to_disk(path=self.ivec_path, Z=Z, F=F,
                                    dtype='float32', device='gpu',
                                    override=True)
      Z.close()
      F.close()
    # ====== clean ====== #
    if not keep_stats:
      if os.path.exists(self.z_path):
        os.remove(self.z_path)
      if os.path.exists(self.f_path):
        os.remove(self.f_path)
    return self

  def transform(self, X, indices=None, sad=None,
                save_ivecs=False, keep_stats=False, name=None):
    """
    Parameters
    ----------
    X : ndarray
      Training data [n_samples, n_features]
    indices : {Mapping, tuple, list}
      in case the data is given by a list of files, `indices`
      act as file indicator mapping from
      'file_name' -> (start_index_in_X, end_index_in_X)
      This mapping can be provided by a dictionary, or list of
      tuple.
    sad : ndarray
      inspired by the "Speech Activity Detection" (SAD) indexing,
      this array is indicator of which samples will be taken into
      training; the shape should be [n_samples,] or [n_samples, 1]
    save_ivecs : bool
      if True, save extracted i-vectors to disk at path `ivec_[name]`
      if False, return directly the i-vectors without saving

    keep_stats : bool
      if True, keep the zero and first order statistics.
      The first order statistics could consume huge amount
      of disk space. Otherwise, they are deleted after training
    name : {None, str}
      identity of the i-vectors (for re-using in future).
      If None, a random name is used
    """
    if not self.is_fitted:
      raise ValueError("Ivector has not been fitted, call Ivector.fit(...) first")
    n_files = X.shape[0] if indices is None else len(indices)
    if name is None:
      name = uuid(length=8)
    else:
      name = str(name)
    # ====== init ====== #
    z_path = self.get_z_path(name)
    f_path = self.get_f_path(name)
    if save_ivecs:
      i_path = self.get_i_path(name)
    else:
      i_path = None
    name_path = self.get_name_path(name)
    # ====== check exist i-vector file ====== #
    if i_path is not None and os.path.exists(i_path):
      ivec = MmapData(path=i_path, read_only=True)
      assert ivec.shape[0] == n_files and ivec.shape[1] == self.tv_dim,\
      "Need i-vectors for %d files, found exists data at path:'%s' with shape:%s" % \
      (n_files, i_path, ivec.shape)
      return ivec
    # ====== extract Z and F ====== #
    if os.path.exists(z_path) and os.path.exists(f_path):
      pass
    else:
      if os.path.exists(z_path):
        os.remove(z_path)
      if os.path.exists(f_path):
        os.remove(f_path)
      if os.path.exists(name_path):
        os.remove(name_path)
      _extract_zero_and_first_stats(X=X, sad=sad, indices=indices, gmm=self.gmm,
                                    z_path=z_path, f_path=f_path, name_path=name_path)
    Z = MmapData(path=z_path, read_only=True)
    F = MmapData(path=f_path, read_only=True)
    # ====== extract I-vec ====== #
    ivec = self.tmat.transform_to_disk(path=i_path, Z=Z, F=F, dtype='float32')
    # ====== clean ====== #
    Z.close()
    F.close()
    if not keep_stats:
      if os.path.exists(z_path):
        os.remove(z_path)
      if os.path.exists(f_path):
        os.remove(f_path)
    else:
      print("Zero-order stats saved at:", ctext(z_path, 'cyan'))
      print("First-order stats saved at:", ctext(f_path, 'cyan'))
    return ivec

  def __str__(self):
    s = ''
    s += ctext("<Ivector ", 'yellow')
    s += "GMM:%s " % self.is_gmm_fitted
    s += "Tmat:%s\n" % self.is_tmat_fitted
    if os.path.exists(self.path) and len(os.listdir(self.path)) > 0:
      # list all model files
      s += "  %s: " % ctext('model', 'cyan')
      s += ', '.join(['"%s"' % f
                      for f in sorted(os.listdir(self.path))
                      if 'zstat' not in f and 'fstat' not in f and
                      'ivec' not in f and 'name_' not in f])
      s += '\n'
      # list all Zero-stats files
      s += "  %s: " % ctext('Z-stats', 'cyan')
      s += ', '.join(['"%s"' % f
                      for f in sorted(os.listdir(self.path))
                      if 'zstat' in f])
      s += '\n'
      # list all First-stats files
      s += "  %s: " % ctext('F-stats', 'cyan')
      s += ', '.join(['"%s"' % f
                      for f in sorted(os.listdir(self.path))
                      if 'fstat' in f])
      s += '\n'
      # list all Ivec-stats files
      s += "  %s: " % ctext('ivec', 'cyan')
      s += ', '.join(['"%s"' % f
                      for f in sorted(os.listdir(self.path))
                      if 'ivec' in f])
      s += '\n'
      # list all Name path files
      s += "  %s: " % ctext('name-list', 'cyan')
      s += ', '.join(['"%s"' % f
                      for f in sorted(os.listdir(self.path))
                      if 'name_' in f])
      s += '\n'
    # list all attributes
    for k, v in sorted(self.__dict__.items(), key=lambda x: x[0]):
      if is_primitives(v, inc_ndarray=False):
        s += "  %s: %s\n" % (ctext(k, 'cyan'), str(v))
    s = s[:-1] + '>'
    return s



from __future__ import print_function, division, absolute_import

import numpy as np
import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin.utils import uuid, as_tuple, flatten_list, is_string, is_number
from odin.autoconfig import get_rng, CONFIG, randint
from odin.backend.helpers import is_training
from odin.backend.tensor import variable
from odin.backend.role import add_roles, Weight, Bias, Parameter

floatX = CONFIG.floatX

def random_binomial(shape, p, dtype=floatX, seed=None, name="RandomBinomal"):
  with tf.variable_scope(name):
    return tf.where(
        tf.random_uniform(shape, minval=0., maxval=1., dtype=dtype, seed=seed) <= p,
        tf.ones(shape, dtype=dtype),
        tf.zeros(shape, dtype=dtype))

# ===========================================================================
# Noise
# ===========================================================================
def _process_noise_dim(input_shape, dims):
  """
  By default, each element is kept or dropped independently.  If `noise_shape`
  is specified, it must be
  [broadcastable](http://docs.scipy.org/doc/numpy/user/basics.broadcasting.html)
  to the shape of `x`, and only dimensions with `noise_shape[i] == shape(x)[i]`
  will make independent decisions.  For example, if `shape(x) = [k, l, m, n]`
  and `noise_shape = [k, 1, 1, n]`, each batch and channel component will be
  kept independently and each row and column will be kept or not kept together.

  Examples
  --------
  (None, 10, 10) with noise_dims=2
  => Noise mask: (None, 10, 1)
  """
  if dims is None:
    return input_shape
  ndims = input_shape.shape[0].value
  dims = [i % ndims for i in as_tuple(dims, t=int)]
  # ====== get noise shape ====== #
  return tuple([1 if i in dims else input_shape[i]
                for i in range(ndims)])


def apply_dropout(x, level=0.5, noise_dims=None, noise_type='uniform',
                  rescale=True, name="ApplyDropout"):
  """Computes dropout.

  With probability `keep_prob`, outputs the input element scaled up by
  `1 / keep_prob`, otherwise outputs `0`.  The scaling is so that the expected
  sum is unchanged.


  Parameters
  ----------
  x: A tensor.
      input tensor
  level: float(0.-1.)
      probability dropout values in given tensor
  noise_dims: int or list(int)
      these dimensions will be setted to 1 in noise_shape, and
      used to broadcast the dropout mask.
  noise_type: 'gaussian' (or 'normal'), 'uniform'
      distribution used for generating noise
  rescale: bool
      whether rescale the outputs by dividing the retain probablity
  seed: random seed or `tensor.rng`
      random generator from tensor class

  References
  ----------
  [Dropout: A Simple Way to Prevent Neural Networks from Overfitting Srivastava, Hinton, et al. 2014](http://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf)

  Note
  ----
  This function only apply noise on Variable when training is enable
  """
  shape = tf.shape(x)
  retain_prob = 1. - level
  # ====== not a training variable NO dropout ====== #
  if 'normal' in noise_type or 'gaussian' in noise_type:
    randfunc = lambda shape: tf.random_normal(shape=shape,
        mean=1.0, stddev=np.sqrt((1.0 - retain_prob) / retain_prob),
        dtype=x.dtype.base_dtype, seed=randint())
  elif 'uniform' in noise_type:
    randfunc = lambda shape: random_binomial(shape=shape,
        p=retain_prob, dtype=x.dtype.base_dtype, seed=randint())
  else:
    raise ValueError('No support for noise_type=' + noise_type)

  # ====== Dropout ====== #
  def training_fn():
    noise_shape = shape if noise_dims is None else \
        _process_noise_dim(shape, noise_dims)
    y = x * randfunc(shape=noise_shape)
    if rescale:
      y /= retain_prob
    return y

  def inference_fn():
    return x
  with tf.variable_scope(name):
    return tf.cond(is_training(), training_fn, inference_fn)


def apply_noise(x, level=0.075, noise_dims=None, noise_type='gaussian',
                name="ApplyNoise"):
  """
  Parameters
  ----------
  x: A tensor.
  level : float or tensor scalar
      Standard deviation of added Gaussian noise
  noise_dims: int or list(int)
      these dimensions will be setted to 1 in noise_shape, and
      used to broadcast the dropout mask.
  noise_type: 'gaussian' (or 'normal'), 'uniform'
      distribution used for generating noise
  seed: random seed or `tensor.rng`
      random generator from tensor class

  Note
  ----
  This function only apply noise on Variable when training is enable
  """
  noise_type = noise_type.lower()
  shape = tf.shape(x)

  # ====== applying noise ====== #
  def training_fn():
    noise_shape = (shape if noise_dims is None
                   else _process_noise_dim(shape, noise_dims))
    if 'normal' in noise_type or 'gaussian' in noise_type:
      noise = tf.random_normal(shape=noise_shape,
          mean=0.0, stddev=level, dtype=x.dtype.base_dtype, seed=randint())
    elif 'uniform' in noise_type:
      noise = tf.random_uniform(shape=noise_shape,
          minval=-level, maxval=level,
          dtype=x.dtype.base_dtype, seed=randint())
    else:
      raise ValueError('No support for noise_type=' + noise_type)
    return x + noise

  # ====== inference_fn ====== #
  def inference_fn():
    return x
  with tf.variable_scope(name):
    return tf.cond(is_training(), training_fn, inference_fn)



"""Annotated computation graph management."""
from __future__ import print_function, absolute_import, division

import re
import warnings
from six import string_types
from contextlib import contextmanager
from collections import OrderedDict, defaultdict, Mapping

import numpy as np

# ====== tensorflow and tensorflow-probability package ====== #
# I am sick of ImportWarning
with warnings.catch_warnings():
  import tensorflow as tf

  _tf_distribution_types = []
  warnings.filterwarnings('ignore', category=ImportWarning)
  try:
    from tensorflow.contrib import distributions as tfd
    _tf_distribution_types.append(tfd.Distribution)
  except ImportError:
    pass
  except AttributeError:
    pass

  try:
    from tensorflow_probability import distributions as _tfd
    _tf_distribution_types.append(_tfd.Distribution)
  except ImportError:
    pass
  except AttributeError:
    pass

  _tf_distribution_types = tuple(_tf_distribution_types)

# ====== O.D.I.N stuffs ====== #
from odin.autoconfig import get_session
from odin.utils.cache_utils import cache_memory
from odin.utils import (dict_union, as_list, flatten_list, as_tuple, is_string,
                        decorators, batching, Progbar)
from odin.backend.role import (has_roles, Auxiliary, Parameter)

# ===========================================================================
# Helper
# ===========================================================================
_TF_SCOPE_PATTERN = lambda scope, begin: \
re.compile(r"%s%s(_\d+)?\/" % ('^' if bool(begin) else '', scope))

# ===========================================================================
# Basic query
# ===========================================================================
def is_training(graph=None):
  if graph is None:
    graph = get_session().graph
  training_var = get_all_variables(scope=None, name='IsTraining__',
                                   graph=graph)
  if len(training_var) == 0:
    raise RuntimeError("Cannot find variable with name='IsTraining' scope='' "
                       "within graph=%s" % str(graph))
  elif len(training_var) > 1:
    raise RuntimeError("Found multiple 'IsTraining__' flag: %s" %
      str(training_var))
  return training_var[0]

def set_training(is_training, graph=None, return_ops=False):
  if graph is None:
    graph = get_session().graph
  return set_value(is_training(graph), bool(is_training),
                   return_ops=return_ops)

def cond_training(train_fn, infer_fn,
                  train_dependencies=None, infer_dependencies=None,
                  strict=False, graph=None,
                  name=None):
  """ Create a conditional output that depends of training variable

  Parameters
  ----------
  train_fn: callable
    this function will be called when `training=True`
  infer_fn: callable
    this function will be called when `training=False`
  train_dependencies: Tensorflow expression
    all of given expression will be executed before running
    the function
  infer_dependencies: Tensorflow expression
    all of given expression will be executed before running
    the function
  strict: bool (default: False)
    if False, automatically group all outputs into single tensorflow
    Tensor, this behaviour is disabled with `strict=True`
  name: str
    specific name for the Operation

  """
  if not hasattr(train_fn, '__call__') or \
  not hasattr(infer_fn, '__call__'):
    raise ValueError("`train_fn` and `infer_fn` must be call-able.")
  with tf.name_scope(name, "TrainingCondition"):
    # ====== training ====== #
    if train_dependencies is not None:
      def _train_fn():
        with tf.control_dependencies(as_tuple(train_dependencies)):
          return train_fn()
    else:
      _train_fn = train_fn
    # ====== inference ====== #
    if infer_dependencies is not None:
      def _infer_fn():
        with tf.control_dependencies(as_tuple(infer_dependencies)):
          return infer_fn()
    else:
      _infer_fn = infer_fn
    return tf.cond(pred=is_training(graph=graph),
                   true_fn=_train_fn, false_fn=_infer_fn,
                   strict=strict)

def is_operation(op):
  return isinstance(op, tf.Operation)

def is_placeholder(variable):
  """Check if variable is a user-provided graph input.

  To be considered an input the variable must have no owner, and not
  be a constant or shared variable.

  Parameters
  ----------
  variable : :class:`~tensor.TensorVariable`

  Returns
  -------
  bool
      ``True`` If the variable is a user-provided input to the graph.

  """
  return isinstance(variable, tf.Tensor) and \
      variable.op.node_def.op == "Placeholder"

def is_variable(variable):
  """Check if a variable is a shared variable.

  Notes
  -----
  This function excludes shared variables that store the state of Theano
  random number generators.

  """
  if (isinstance(variable, tf.Variable) and
          variable.op.node_def.op == "VariableV2"):
    return True
  return False

def is_distribution(x):
  return isinstance(x, _tf_distribution_types)

def is_tensor(variable, inc_distribution=True, inc_variable=True):
  """ a variable is any tensor variable in (e.g. placeholder,
  trainable_variable, intermediate tensor, ...)
  All `TensorType` includes:
  * ops.Tensor
  * sparse_tensor.SparseTensor
  * variables.Variable
  """
  _ = tf.contrib.framework.is_tensor(variable)
  if not inc_variable:
    _ &= (not isinstance(variable, tf.Variable))
  if inc_distribution:
    _ |= is_distribution(variable)
  return _

def set_shape(tensor, shape):
  """ This function will filling the missing shape information
  of given tensor
  """
  if not is_tensor(tensor):
    raise ValueError('tensor must be instance of `Tensor`.')
  # ====== Test ====== #
  ndims = tensor.shape.ndims
  shape = as_tuple(shape)
  if ndims != len(shape):
    raise ValueError("The tensor has %d dimensions, but the given shape "
                     "has %d dimension." % (ndims, len(shape)))
  # ====== DO it ====== #
  old_shape = tensor.shape
  new_shape = []
  for old, new in zip(old_shape, shape):
    old_value = old.value
    if isinstance(new, tf.Dimension):
      new = new.value
    # matching old and new values
    if old_value is not None and new is not None:
      if old_value != new:
        raise ValueError("Known shape information mismatch, from tensorflow"
            ":%s, and given shape:%s." %
            (str(old_shape.as_list()), str(shape)))
      else:
        new_shape.append(old_value)
    elif old_value is None and new is not None:
      new_shape.append(new)
    elif old_value is not None and new is None:
      new_shape.append(old_value)
    elif old is None and new is None:
      new_shape.append(old)
    else:
      new_shape.append(None)
  tensor.set_shape(new_shape)
  return tensor

# ===========================================================================
# VALUE MANIPULATION
# ===========================================================================
def get_value(x):
  if isinstance(x, (tuple, list)):
    return get_session().run(x)
  return x.eval(session=get_session())

def set_value(x, value, return_ops=False, name='SetValue'):
  '''Sets the value of a tensor variable,
  from a Numpy array.

  Parameters
  ----------
  x: `Tensor`
  value: real value
  return_ops: bool
      if True, return assign Op and feed_dict instead of running
      the Op directly
  '''
  if isinstance(value, np.ndarray):
    value = value.astype(x.dtype.as_numpy_dtype)
  elif is_tensor(value):
    value = tf.cast(value, dtype=x.dtype)
  assign_op = tf.assign(x, value, name=name)
  if return_ops:
    return assign_op
  get_session().run(assign_op)
  return x

def get_epsilon(x_or_dtype):
  """ Return the epsilon given a dtype or dtype of given Tensor

  Parameters
  ----------
  x_or_dtype : {numpy.ndarray, tensorflow.Tensor, numpy.dtype, string}
      anything that contain the given dtype information
  """
  if isinstance(x_or_dtype, string_types):
    dtype = np.dtype(x_or_dtype)
  elif hasattr(x_or_dtype, 'dtype'):
    dtype = x_or_dtype.dtype
    if hasattr(dtype, 'as_numpy_dtype'):
      dtype = dtype.as_numpy_dtype
  else:
    raise ValueError("Cannot infer dtype from: %s" %
      str(type(x_or_dtype)))
  return np.finfo(dtype).eps

# ===========================================================================
# Session helper
# ===========================================================================
def _filter_string(criterion, x):
  if isinstance(criterion, string_types):
    return criterion == x
  elif hasattr(criterion, '__call__'):
    return criterion(x)
  elif criterion is None:
    return True
  raise ValueError("Unknown criterion for filtering.")

_ops_ID = {}
_name_pattern = re.compile(r".*_\d+")

def get_normalized_name(var, shape=True, dtype=True):
  """Get normalized name of `Tensor`, `Variable` or `Op` in tensorflow
  The normalized name remove:
   * '_%d' suffix of scope and name.
   * ':%d' indices of name.

  shape: if True, append string of shape to name
  dtype: if True, append string of dtype to name
  """
  x = var.name
  # remove the index
  x = x.split(':')[0]
  # remove the _%d
  x = '/'.join(['_'.join(i.split('_')[:-1]) if _name_pattern.match(i) else i
                for i in x.split('/')])
  if shape:
    x += str(tuple(var.shape.as_list())).replace(' ', '')
  if dtype:
    x += np.dtype(var.dtype.as_numpy_dtype).name
  return x

def get_all_operations(otype=None, device=None, sort=False, scope=None,
                       footprint=None, graph=None, beginning_scope=True):
  """ Return list of all operations in default graph
  The follow attributes can be access within the operation:
   * name : string
   * otype : string, operation type (e.g. `"MatMul"`).
   * device:  string name of the device to which this op has been assigned
   * _inputs : list of `Tensor`
   * _outputs : list of `Tensor`
   * _control_inputs : Before this op is executed, the operations in
       `control_inputs` have finished executing.
   * graph : `Graph` that contains this operation
   * node_def : serialized `NodeDef` representation of this operation.
   * op_def : `OpDef` proto that represents the type of this op.
   * traceback : call stack from when this operation was constructed.

  Some important op type:
   * "Placeholder"
   * "VariableV2"
   * "Const"
   * "Assign"

  Parameters
  ----------
  beginning_scope : bool (default: True)
    if True, the provide scope must be the beginning scope,
    otherwise, it could be in the middle of multiple scopes
  """
  if graph is None:
    graph = get_session().graph
  ops = graph.get_operations()
  # update OpID
  if len(_ops_ID) != len(ops):
    for ID, op in graph._nodes_by_id.items():
      if op not in _ops_ID:
        _ops_ID[op] = ID
  # filter out some op
  if otype is not None:
    ops = [o for o in ops
           if _filter_string(otype, o.type)]
  if device is not None:
    ops = [o for o in ops
           if _filter_string(device, o.device)]
  # ====== filter by scope ====== #
  if scope is not None:
    scope = str(scope)
    if len(scope) == 0:
      ops = [o for o in ops
             if '/' not in o.name]
    else:
      scope_name_pattern = _TF_SCOPE_PATTERN(scope, beginning_scope)
      ops = [o for o in ops
             if len(scope_name_pattern.findall(o.name))]
  # ====== filter by unique footprint ====== #
  if footprint is not None:
    ops = [o for o in ops
           if get_operation_footprint(o) == footprint]
  # sorted by OpID
  if sort and len(ops) > 1:
    ops = sorted(ops, key=lambda x: _ops_ID[x])
  return ops

def get_operationID(op, graph=None):
  """operation ID is unique ID of Op, the ID represent the order
  of created Op."""
  if graph is None:
    graph = get_session().graph
  ops = graph.get_operations()
  # update OpID
  if len(_ops_ID) != len(ops):
    for ID, op in graph._nodes_by_id.items():
      if op not in _ops_ID:
        _ops_ID[op] = ID
  return _ops_ID[op]

@cache_memory
def get_operation_footprint(op):
  """ Trace back the inputs of given Op and record all:
  * placholders
  * variables
  * ops
  Those are related to given op.

  The final footprint is concatenated string of all variables,
  placeholders, constants, and Ops

  Note
  ----
  This is just a fair attempt to create short identification of a
  tenorflow Op
  still have chance for duplication op footprint
  """
  if not isinstance(op, tf.Operation) and hasattr(op, 'op'):
    op = op.op
  op_id = str(op.name) + str(op.type)
  var = []
  placeholder = []
  const = []
  ops = [op.type]
  inputs = list(op.inputs)
  while len(inputs) > 0:
    i = inputs.pop()
    o = i.op
    ops.append(o.type)
    if o.type == "VariableV2":
      var.append(i)
    elif o.type == "Placeholder":
      placeholder.append(i)
    elif o.type == "Const":
      const.append(i)
    inputs = list(o.inputs) + inputs
  return op_id + '|' +\
  ':'.join([get_normalized_name(v) for v in var]) + '|' +\
  ':'.join([get_normalized_name(p) for p in placeholder]) + '|' +\
  ':'.join([get_normalized_name(c) for c in const]) + '|' +\
  ':'.join([j.split(':')[0] for j in ops])

def get_all_variables(scope=None, name=None, full_name=None,
                      graph_keys=[tf.GraphKeys.GLOBAL_VARIABLES,
                                  tf.GraphKeys.LOCAL_VARIABLES,
                                  tf.GraphKeys.MODEL_VARIABLES,
                                  tf.GraphKeys.TRAINABLE_VARIABLES],
                      graph=None, beginning_scope=True):
  """
  Parameters
  ----------
  scope: {str, None}
      scope name which the Variables have been created
  name: str
      name of tensor (WITHOUT variable scope)
  full_name: str
      name of tensor WITH variable scope.
  beginning_scope : bool (default: True)
    if True, the provide scope must be the beginning scope,
    otherwise, it could be in the middle of multiple scopes
  """
  var = []
  # ====== first get all available variable ====== #
  for k in graph_keys:
    if graph is None:
      var += [i for i in tf.get_collection(k)
              if isinstance(i, tf.Variable)]
    else:
      var += [i for i in graph.get_collection(k)
              if isinstance(i, tf.Variable)]
  var = list(set(var))
  # filtering: start from general to detail
  # ====== filter by scope ====== #
  if scope is not None:
    scope = str(scope)
    if len(scope) == 0:
      var = [v for v in var
             if '/' not in v.name]
    else:
      scope_name_pattern = _TF_SCOPE_PATTERN(scope, beginning_scope)
      var = [v for v in var
             if len(scope_name_pattern.findall(v.name))]
  # ====== filter by name ====== #
  if name is not None:
    name = as_tuple(name, t=string_types)
    var = [v for v in var
           if any((v.name.split('/')[-1] == n or
                   v.name.split('/')[-1].split(':')[0] == n.split(':')[0])
                  for n in name)]
  # ====== filter by fullname ====== #
  if full_name is not None:
    full_name = as_tuple(full_name, t=string_types)
    var = [v for v in var
           if any((n == v.name or
                   n.split(':')[0] == v.name.split(':')[0])
                  for n in full_name)]
  return var

def get_all_tensors(scope=None, name=None, full_name=None, device=None):
  """
  Parameters
  ----------
  scope: {str, None}
      scope name which the Variables have been created
  name: str
      name of tensor (without variable scope)
  full_name: str
      name of tensor WITH variable scope.
  device : {str, None}
      name of the device to which this op has been assigned
      (e.g. /cpu:0, or /gpu:0)
  """
  ops = get_all_operations(device=device, scope=scope, sort=False)
  alltensors = []
  for o in ops:
    alltensors += list(o.inputs) + list(o._outputs)
    for i in o.control_inputs:
      alltensors += list(i.inputs) + list(i._outputs)
  alltensors = list(set(alltensors))
  # ====== filter out unsupport types ====== #
  if name is not None:
    name = as_tuple(name, t=string_types)
    alltensors = [t for t in alltensors
                  if any((n == t.name.split('/')[-1] or
                          n.split(':')[0] == t.name.split('/')[-1].split(':')[0])
                         for n in name)]
  if full_name is not None:
    full_name = as_tuple(full_name, t=string_types)
    alltensors = [t for t in alltensors
                  if any((n == t.name or
                          n.split(':')[0] == t.name.split(':')[0])
                        for n in full_name)]
  return alltensors

def get_all_variables_or_tensors(scope=None, name=None, full_name=None):
  var = get_all_variables(scope=scope, name=name, full_name=full_name)
  if len(var) == 0:
    var = get_all_tensors(scope=scope, name=name, full_name=full_name)
  return var

# ===========================================================================
# ComputationGraph
# ===========================================================================
class Function(object):
  """ Two way to call this Function
  f(x1, x2, x3) or f('x1'=x1, 'x2'=x2, 'x3'=x3)

  Parameters
  ----------
  inputs: list of `tf.placeholder` or `tf.Variable`

  outputs: list of `tf.Tensor`

  updates: list, or dict
      mapping from `Tensor` to its new value which is `Tensor` or
      real value.

  defaults: dict
      mapping from `Variable` or `placeholder` to its default values.

  training: None, True, False
      if `training=None`, left the training mode unchanged
      if `training=True`, turn on training mode only when execute this
      function.
      if `training=False`, disable training mode only when execute this
      function.

  batch_size : {int, None} (default: None)
      if `batch_size` is not None, auto-split all array into minibatch,
      and return a list of outputs (all array must have equal `shape[0]`)

  batch_vars : {Tensor, list of Tensor}
      if `len(batch_vars) == 0`, split mini-batches for all Tensor inputs,
      otherwise, only applying for a selected set of inputs.

  strict : bool (default: True)
      if False, remove shape mis-matched inputs when `__call__`
      this `Function`.
      if True, raise RuntimeError.

  Note
  ----
  call the function with `show_progress=True` when `batch_size`
  is specified to show the progress.

  """

  def __init__(self, inputs, outputs, updates=[], defaults={},
               training=None, batch_size=None, batch_vars=[],
               strict=False):
    self.training = training
    self.batch_size = batch_size
    self.batch_vars = list(batch_vars)
    self._strict = bool(strict)
    # ====== validate input ====== #
    if isinstance(inputs, Mapping):
      self.inputs_name = inputs.keys()
      inputs = inputs.values()
    elif not isinstance(inputs, (tuple, list)):
      inputs = [inputs]
    self.inputs = flatten_list(inputs, level=None)
    if not hasattr(self, 'inputs_name'):
      self.inputs_name = [i.name.split(':')[0] for i in self.inputs]
    # ====== defaults ====== #
    defaults = dict(defaults)
    self.defaults = defaults
    # ====== validate outputs ====== #
    return_list = True
    if not isinstance(outputs, (tuple, list)):
      outputs = (outputs,)
      return_list = False
    self.outputs = flatten_list(list(outputs), level=None)
    self._return_list = return_list
    # ====== validate updates ====== #
    if isinstance(updates, Mapping):
      updates = updates.items()
    with tf.control_dependencies(self.outputs):
      # create updates ops
      if not isinstance(updates, tf.Operation):
        updates_ops = []
        for update in updates:
          if isinstance(update, (tuple, list)):
            p, new_p = update
            updates_ops.append(tf.assign(p, new_p))
          else: # assumed already an assign op
            updates_ops.append(update)
        self.updates_ops = tf.group(*updates_ops)
      else: # already an tensorflow Ops
        self.updates_ops = updates
    # ====== cached shape ====== #
    self._input_shape = [tuple(i.shape.as_list()) for i in self.inputs]
    self._output_shape = [tuple(i.shape.as_list()) for i in self.outputs]

  @property
  def input_shape(self):
    return self._input_shape

  @property
  def output_shape(self):
    return self._output_shape if self._return_list else self._output_shape[0]

  def __call__(self, *inputs, **kwargs):
    show_progress = kwargs.pop('show_progress', False)
    # dictionary as inputs
    if len(kwargs) == len(self.inputs_name):
      inputs = [kwargs[i] for i in self.inputs_name]
    # ====== delete un-matchede inputs ====== #
    inputs_new = []
    tmp = list(inputs)
    shapes = list(self._input_shape)
    # this process iteratively remove inputs with mismatch shape
    # to current given input
    for s in shapes:
      for i in tuple(tmp):
        if len(i.shape) != len(s) or \
        any(a is not None and a > 0 and a != b
                for a, b in zip(s, i.shape)): # different ndim, or shape
          tmp.remove(i)
        else:
          inputs_new.append(i)
          tmp.remove(i)
          break
    if len(inputs_new) != len(self.inputs):
      raise ValueError("Given inputs have shape: %s, cannot match the shape of "
                       "defined inputs: %s" %
                       ('; '.join([str(i.shape) for i in inputs]),
                        '; '.join([str(i) for i in self.input_shape])))
    if not self._strict:
      inputs = inputs_new
    # ====== create feed_dict ====== #
    feed_dict = {}
    inputs = flatten_list(inputs, level=None)
    for tensor, value in zip(self.inputs, inputs):
      feed_dict[tensor] = value
    feed_dict.update(self.defaults)
    # check if modifying training mode
    if self.training is None:
      pass
    elif self.training:
      feed_dict.update({is_training(): True})
    else:
      feed_dict.update({is_training(): False})
    session = get_session()
    outputs = None
    # ====== mini-batches ====== #
    if self.batch_size is not None:
      batch_vars = ([i for i in feed_dict.keys() if is_tensor(i)]
                    if len(self.batch_vars) == 0 else self.batch_vars)
      batch_vars = [i for i in batch_vars
                    if i in feed_dict and hasattr(feed_dict[i], 'shape')]
      n_samples = list(set(feed_dict[i].shape[0] for i in batch_vars))
      assert len(n_samples) == 1, \
      "Data have multiple batching dimension: %s" % str(n_samples)
      n_samples = n_samples[0]
      # only continue if we have more samples than `batch_size`
      if n_samples > self.batch_size:
        n_output = len(self.outputs)
        outputs = []
        all_batches = []
        # (optional) showing progress
        if show_progress:
          prog = Progbar(target=n_samples,
                         print_report=False, print_summary=False,
                         name='')
        for s, e in batching(batch_size=int(self.batch_size),
                             n=n_samples):
          if show_progress:
            prog.add(e - s)
          all_batches.append(e - s)
          feed_dict_minibatch = OrderedDict([(k, v[s:e])
                                             if k in batch_vars else (k, v)
                                             for k, v in feed_dict.items()])
          updated = session.run(self.outputs + [self.updates_ops],
                                feed_dict=feed_dict_minibatch)
          updated = updated[:n_output]
          if not self._return_list:
            updated = updated[0]
          outputs.append(updated)
        ## concatenate all outputs
        if not self._return_list:
          o_ndim = outputs[0].ndim
          if o_ndim == 0: # returned scalars
            outputs = np.array(outputs)
          else: # returned array
            for o_axis in range(o_ndim):
              all_n = [o.shape[o_axis] for o in outputs]
              if all_n == all_batches:
                break
            outputs = np.concatenate(outputs, axis=o_axis)
        ## returning a list of outputs
        else:
          new_outputs = []
          for output_idx in range(len(outputs[0])):
            o = [x[output_idx] for x in outputs]
            o_ndim = o[0].ndim
            if o_ndim == 0: # returned scalars
              o = np.array(o)
            else: # returned array
              for o_axis in range(o[0].ndim):
                all_n = [val.shape[o_axis] for val in o]
                if all_n == all_batches:
                  break
              o = np.concatenate(o, axis=o_axis)
            new_outputs.append(o)
          outputs = new_outputs
    # ====== single batch ====== #
    if outputs is None:
      updated = session.run(self.outputs + [self.updates_ops],
                            feed_dict=feed_dict)
      outputs = updated[:len(self.outputs)]
      if not self._return_list:
        outputs = outputs[0]
    # ====== return final output ====== #
    return outputs

def function(inputs, outputs, updates=[], defaults={},
             training=None, batch_size=None, batch_vars=[]):
  """
  Parameters
  ----------
  inputs: list of `tf.placeholder` or `tf.Variable`
  outputs: list of `tf.Tensor`
  updates: list, or dict
      mapping from `Tensor` to its new value which is `Tensor` or
      real value.
  defaults: dict
      mapping from `Variable` or `placeholder` to its default values.
  training: None, True, False
      if `training=None`, left the training mode unchanged
      if `training=True`, turn on training mode only when execute this
      function.
      if `training=False`, disable training mode only when execute this
      function.
  batch_size : {int, None} (default: None)
      if `batch_size` is not None, auto-split all array into minibatch,
      and return a list of outputs (all array must have equal `shape[0]`)
  batch_vars : {Tensor, list of Tensor}
      if `len(batch_vars) == 0`, split mini-batches for all Tensor inputs,
      otherwise, only applying for a selected set of inputs.
  """
  # ====== check inputs ====== #
  if inputs is None or len(as_tuple(inputs)) == 0:
    inputs = ComputationGraph(outputs).placeholders
    inputs_text = ', '.join([str(i) for i in inputs])
    print("[WARNING] inputs haven't specified, auto-inferred from Graph of "
          "outputs, graph inputs: %s" % 'None' if len(inputs) == 0 else inputs_text)
  return Function(inputs=inputs, outputs=outputs,
                  updates=updates, defaults=defaults,
                  training=training, batch_size=batch_size)

# ===========================================================================
# Computational graph
# ===========================================================================
def get_intermediate_tensors(outputs, scope=None, name=None, full_name=None,
                             roles=None, match_all=False, exact=False,
                             beginning_scope=False):
  """ Return all variables and tensor within the computational
  graph of the `outputs` that are satisfied the given conditions

  Parameters
  ----------
  scope : {None, string}
    name of variable scope, any of the scope that match given name
    will be selected
  name : {None, string}
    the name of tensor without the output indexing ":0" and the scope
  full_name : {None, string}
    the full name includes both scope and tensor name without the
    output indexing ":0"
  roles : {None, odin.backend.role}
    specific roles of the tensor
  match_all : bool (default: False)
    If ``True``, checks if the variable has all given roles.
    If ``False``, any of the roles is sufficient.
  exact : bool (default: False)
    If ``True``, use ``==`` for comparison to get exactly same roles.
    If ``False``, use `issubclass` for comparison, hence, also match the
    descendant roles.
  beginning_scope : bool (default: False)
    if True, the provide scope must be the beginning scope,
    otherwise, it could be in the middle of multiple scopes
  """
  graph = ComputationGraph(outputs)
  return graph.get(scope=scope, name=name, full_name=full_name,
                   roles=roles, match_all=match_all, exact=exact,
                   beginning_scope=beginning_scope)

@decorators.singleton
class ComputationGraph(object):
  r"""Encapsulates a managed Theano computation graph.

  This implies that it not only contains the variables required to
  compute the given outputs, but also all the auxiliary variables and
  updates that were attached to these variables through the annotation
  system.

  All variables are presented in topologically sorted order according to
  the apply nodes that they are an input to.

  Parameters
  ----------
  outputs : (list of) :class:`~tf.Tensor`
      The output(s) of the computation graph.
  trace_up : bool
      if True, all the descendant `Tensor` that computed based on
      those outputs will be traced.

  Note
  ----
  This :class:`ComputationGraph` is a `singleton` class since once a graph
  is created, it will never be changed (node insertion or deletion).

  Attributes
  ----------
  inputs : list of :class:`~tensor.TensorVariable`
      The inputs of the computation graph. This does not include shared
      variables and constants.
  trainable_variables : list of :class:`~tensor.TensorSharedVariable`
      All the shared variables in the graph.
  parameters : list of :class:`~tensor.TensorSharedVariable`
      All the shared variables which have the :const:`.PARAMETER` role.
  outputs : list of :class:`~tensor.TensorVariable`
      The outputs of the computations graph (as passed to the
      constructor).
  auxiliary_variables : list of :class:`~tensor.TensorVariable`
      All variables which have the :const:`.AUXILIARY` role.
  intermediary_variables : list of :class:`~tensor.TensorVariable`
      Any variable that is not part of :attr:`inputs` or :attr:`outputs`.
  variables : list of :class:`~tensor.TensorVariable`
      All variables (including auxiliary) in the managed graph.
  scans : list of :class:`~theano.scan_module.scan_op.Scan`
      All Scan ops used in this computation graph.
  scan_variables : list of :class:`~tensor.TensorVariable`
      All variables of the inner graphs of Scan ops.
  updates : :class:`~tensor.TensorSharedVariable` updates
      All the updates found attached to the annotations.

  """

  @classmethod
  def _get_id(clazz, outputs=None, trace_up=False):
    return outputs

  def __init__(self, outputs=None, trace_up=False):
    # it is important to don't have duplicated outputs
    # otherwise, it can go into infinite loop
    outputs = list(set([o for o in flatten_list(as_list(outputs),
                                           level=None)
                        if o is not None]))
    self.outputs = outputs
    self._trace_up = trace_up
    self._get_variables()

  def _get_variables(self):
    """ Collect variables, updates and auxiliary variables.

    In addition collects all :class:`.Scan` ops and recurses in the
    respective inner Theano graphs. """
    _travelled_up = [] # to prevent recursive ops
    _travelled_down = [] # to prevent recursive ops

    def get_all_tensor_trace_down(x):
      """ recursively travel down the inputs tree to get all
      tensor """
      tensors = []
      op = x.op
      # ====== check travelled ops ====== #
      if op in _travelled_down:
        return variables
      else:
        _travelled_down.append(op)
      # ====== get all variable ====== #
      inputs = list(op.inputs)
      tensors += inputs
      for i in inputs:
        tensors += get_all_tensor_trace_down(i)
      return tensors

    def get_all_tensor_trace_up(x):
      """ travel up the outputs tree to get all variables"""
      tensors = []
      # ====== check travelled ops ====== #
      for op in x.consumers():
        if op in _travelled_up:
          continue
        else:
          _travelled_up.append(op)
        # ====== get all variable ====== #
        inputs = [i for i in op.inputs if i != x]
        outputs = list(op._outputs)
        tensors += inputs + outputs
        for o in outputs:
          tensors += get_all_tensor_trace_up(o)
      return tensors

    def create_tensor_iter(outputs):
      # if specific outputs is given
      if len(outputs) > 0:
        for o in outputs:
          # travese each node of graph
          all_tensors = get_all_tensor_trace_down(o)
          if self._trace_up:
            all_tensors += get_all_tensor_trace_up(o)
          for t in all_tensors:
            yield t
      # get all variables and tensor within the graph
      else:
        for o in outputs:
          with o.graph.as_default():
            for op in get_all_operations(sort=False):
              for t in list(op.inputs) + list(op._outputs):
                yield t
    # store all the updates embedded into the Tensor Variables
    variables = [o for o in self.outputs if is_variable(o)]
    outputs = [o for o in self.outputs if not is_variable(o)]
    tensors = []
    placeholders = []
    # ====== travese each node of graph ====== #
    # first get all available variables
    # (even not related to outputs)
    global_vars = {}
    if len(outputs) > 0:
      for o in outputs:
        with o.graph.as_default():
          for v in get_all_variables():
            global_vars[v.name] = v
            global_vars[v.value().name] = v
    else:
      for v in get_all_variables():
        global_vars[v.name] = v
        global_vars[v.value().name] = v
    # then iterate over all tensor
    for t in create_tensor_iter(outputs):
      if t.name in global_vars:
        variables.append(global_vars[t.name])
      elif is_tensor(t):
        tensors.append(t)
    variables = list(set(variables))
    tensors = list(set(tensors + outputs))
    # sorted by Ops ID in _nodes=
    graph_nodes_ID = {}
    for t in tensors:
      if t.graph not in graph_nodes_ID:
        graph_nodes_ID[t.graph] = {op: ID
            for ID, op in t.graph._nodes_by_id.items()}
    tensors = sorted(tensors, key=lambda x: graph_nodes_ID[x.graph][x.op])
    placeholders = [t for t in tensors if is_placeholder(t)]

    self._placeholders = placeholders
    self._tensors = tensors
    self._variables = variables

  # ==================== Get variables ==================== #
  @property
  def placeholders(self):
    """Inputs to the graph, excluding constants and shared variables."""
    return list(self._placeholders)

  @property
  def tensors(self):
    return self._tensors

  @property
  def variables(self):
    return list(self._variables)

  @property
  def parameters(self):
    return [var for var in self._variables if has_roles(var, Parameter)]

  @property
  def auxiliary_tensors(self):
    return [t for t in self._tensors if has_roles(t, Auxiliary)]

  @property
  def dict_of_placeholders(self):
    """Return a mapping from an input name to the input."""
    return {var.name.split(':')[0]: var for var in self.placeholders}

  # ==================== Graph manipulation ==================== #
  def swap(self, dict_swap, scope="copied", replace_itself=False, copy_q=False):
    """ Original implementation: Edward (https://github.com/blei-lab/edward)
    Build a new node in the TensorFlow graph from `org_instance`,
    where any of its ancestors existing in `dict_swap` are
    replaced with `dict_swap`'s corresponding value.

    The copying is done recursively, so any `Operation` whose output
    is required to evaluate `org_instance` is also copied (if it isn't
    already copied within the new scope). This is with the exception of
    `tf.Variable`s, `tf.placeholder`s, and nodes of type `Queue`, which
    are reused and not newly copied.

    Parameters
    ----------
    dict_swap : dict, optional
      Random variables, variables, tensors, or operations to swap with.
      Its keys are what `org_instance` may depend on, and its values are
      the corresponding object (not necessarily of the same class
      instance, but must have the same type, e.g., float32) that is used
      in exchange.
    scope : str, optional
      A scope for the new node(s). This is used to avoid name
      conflicts with the original node(s).
    replace_itself : bool, optional
      Whether to replace `org_instance` itself if it exists in
      `dict_swap`. (This is used for the recursion.)
    copy_q : bool, optional
      Whether to copy the replaced tensors too (if not already
      copied within the new scope). Otherwise will reuse them.

    Returns
    -------
    RandomVariable, tf.Variable, tf.Tensor, or tf.Operation
      The copied node.

    Examples
    --------
    >>> from odin import backend as K
    >>> x = tf.constant(2.0, name='x')
    >>> y = tf.constant(3.0, name='y')
    >>> z = tf.multiply(x, y, name="z")
    ... # define replacement variables
    >>> qx = tf.constant(4.0, name='qx')
    >>> qz = tf.constant(25.0, name='qz')
    ... # The TensorFlow graph is currently
    ... # `x` -> `z` <- y`, `qx`
    ... # This adds a subgraph with newly copied nodes,
    ... # `copied/qx` -> `copied/z` <- `copied/y`
    >>> z_new = K.ComputationGraph(z).swap(
    ...     dict_swap={x: qx, z: qz},
    ...     replace_itself=False, copy_q=False)
    >>> print([v.name for v in K.ComputationGraph(z_new).variables])
    ... # [u'qx:0', u'copied/y:0', u'copied/z:0', u'copied/w:0']
    >>> sess = tf.Session()
    >>> sess.run(z) # 6.0
    >>> sess.run(z_new) # 12.0
    ... # with replace_itself = True
    >>> z_new = K.ComputationGraph(z).swap(
    ...     dict_swap={x: qx, z: qz},
    ...     replace_itself=True, copy_q=False)
    >>> print([v.name for v in K.ComputationGraph(z_new).variables])
    ... # [u'qx:0', u'copied/y:0', u'qz:0', u'copied/w:0']
    >>> sess.run(z_new) # 25.0
    """
    try:
      from edward import copy
    except ImportError:
      raise RuntimeError("Require Edward library to manipulate the "
                         "ComputationGraph.")
    outputs_new = []
    for o in self.outputs:
      o_new = copy(org_instance=o, dict_swap=dict_swap, scope=scope,
               replace_itself=replace_itself, copy_q=copy_q)
      dict_swap[o] = o_new
      outputs_new.append(o_new)
    return outputs_new

  def get(self, scope=None, name=None, full_name=None,
          roles=None, match_all=False, exact=False,
          beginning_scope=False):
    """ Return all variables and tensor with given roles

    Parameters
    ----------
    scope : {None, string}
      name of variable scope, any of the scope that match given name
      will be selected
    name : {None, string}
      the name of tensor without the output indexing ":0" and the scope
    full_name : {None, string}
      the full name includes both scope and tensor name without the
      output indexing ":0"
    roles : {None, odin.backend.role}
      specific roles of the tensor
    match_all : bool (default: False)
      If ``True``, checks if the variable has all given roles.
      If ``False``, any of the roles is sufficient.
    exact : bool (default: False)
      If ``True``, use ``==`` for comparison to get exactly same roles.
      If ``False``, use `issubclass` for comparison, hence, also match the
      descendant roles.
    beginning_scope : bool (default: True)
      if True, the provide scope must be the beginning scope,
      otherwise, it could be in the middle of multiple scopes
    """
    alltensors = self.tensors + self.variables
    # ====== by role ====== #
    if roles is not None:
      alltensors = [t for t in alltensors
                    if has_roles(t, roles=roles,
                                 match_all=match_all, exact=exact)]
    # ====== from general to detail ====== #
    if scope is not None:
      scope = str(scope)
      if len(scope) == 0:
        alltensors = [t for t in alltensors
                      if '/' not in t.name]
      else:
        scope_name_pattern = _TF_SCOPE_PATTERN(scope, beginning_scope)
        alltensors = [t for t in alltensors
                      if len(scope_name_pattern.findall(t.name))]
    # ====== filter by name ====== #
    if name is not None:
      name = as_tuple(name, t=string_types)
      alltensors = [t for t in alltensors
                    if any((n == t.name.split('/')[-1] or
                            n.split(':')[0] == t.name.split('/')[-1].split(':')[0])
                           for n in name)]
    # ====== full name ====== #
    if full_name is not None:
      full_name = as_tuple(full_name, t=string_types)
      alltensors = [t for t in alltensors
                    if any((n == t.name or
                            n.split(':')[0] == t.name.split(':')[0])
                           for n in full_name)]
    return alltensors

  # ==================== others ==================== #
  def __len__(self):
    return len(self.variables)

  def __iter__(self):
    for v in self.variables:
      yield v

  def __del__(self):
    self.dispose()
    del self.outputs
    del self.variables



# -*- coding: utf-8 -*-
# ===========================================================================
# Popular encoding:
#  utf-8
#  ISO-8859-1
#  ascii
# encode: string -> string of bytes
# decode: string of bytes -> string
# ===========================================================================
from __future__ import print_function, division, absolute_import

import timeit
import string
from collections import OrderedDict, Iterator, Iterable, defaultdict, Mapping
from abc import abstractmethod, ABCMeta
from six import add_metaclass, string_types

import numpy as np

from odin.utils import as_tuple, Progbar, is_string, is_number
from odin.stats import freqcount
from multiprocessing import Pool, cpu_count

from odin.preprocessing.signal import pad_sequences

_nlp = {}
_stopword_list = []


# ===========================================================================
# Helper
# ===========================================================================
def language(lang='en'):
  """Support language: 'en', 'de' """
  import spacy
  lang = lang.lower()
  if lang not in ('en', 'de'):
    raise ValueError('We only support languages: en-English, de-German.')
  if lang not in _nlp:
    if lang == 'en':
      # "en_core_web_md"
      _nlp[lang] = spacy.load(lang)
  return _nlp[lang]


def add_stopword(words):
  words = as_tuple(words, t=string_types)
  for w in words:
    _stopword_list.append(w)


def is_stopword(word, lang='en'):
  nlp = language(lang)
  # ====== check self-defined dictionary ====== #
  if word in _stopword_list:
    return True
  # ====== check in spacy dictionary ====== #
  if word not in nlp.vocab.strings:
    return False
  lexeme = nlp.vocab[nlp.vocab.strings[word]]
  return lexeme.is_stop


def is_oov(word, lang='en'):
  """ Check if a word is out of dictionary """
  nlp = language(lang)
  if word not in nlp.vocab.strings:
    return True
  return False


# ===========================================================================
# Text preprocessor
# ===========================================================================
@add_metaclass(ABCMeta)
class TextPreprocessor(object):
  """ A Preprocessor takes a string and return a preprocessed string
  a list of strings which represented a list of tokens.
  """

  @abstractmethod
  def preprocess(self, text):
    pass

  def __call__(self, text):
    if isinstance(text, (tuple, list)):
      return [self.preprocess(text) for t in text]
    else:
      return self.preprocess(text)


class CasePreprocessor(TextPreprocessor):
  """Remove case, name, and split the text"""

  def __init__(self, lower, keep_name=True, split=' '):
    super(CasePreprocessor, self).__init__()
    self.lower = bool(lower)
    self.split = split
    self.keep_name = keep_name

  def preprocess(self, text):
    if self.split is not None:
      text = text.split(' ')
      if self.lower:
        text = [t if self.keep_name and t.isupper()
                else t.lower()
                for t in text if len(t) > 0]
    elif self.lower:
      text = text.lower()
    return text


class TransPreprocessor(TextPreprocessor):
  """ Substitute a set of character to a new characters """

  def __init__(self, old='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n',
               new=' '):
    super(TransPreprocessor, self).__init__()
    self.__str_trans = string.maketrans(old, new * len(old))
    new = None if len(new) == 0 else unicode(new)
    old = unicode(old)
    self.__uni_trans = dict((ord(char), new) for char in old)

  def preprocess(self, text):
    if isinstance(text, (tuple, list)):
      text = ' '.join(text)
    # ====== translate the text ====== #
    if isinstance(text, unicode):
      text = text.translate(self.__uni_trans)
    else:
      text = text.translate(self.__str_trans)
    return text.strip()


# ===========================================================================
# Simple Token filter
# ===========================================================================
@add_metaclass(ABCMeta)
class TokenFilter(object):
  """ A Filter return a "length > 0" token if the token is accepted
  and '' token otherwise.

  This allows not only token filtering but also token transforming.
  """

  @abstractmethod
  def filter(self, token, pos, is_stop, is_oov):
    pass

  def __call__(self, token, pos):
    return self.filter(token, pos)


class TYPEfilter(TokenFilter):
  """ Simplest form of filter
  is_alpha: alphabetic characters only (i.e. no space and digit)
  is_digit: consists of digits only
  is_ascii: all character must be ASCII
  is_title: the first letter if upper-case
  is_punct: punctuation

  * if any value True, the token with given type is accepted

  Paramters
  ---------
  keep_oov: bool
      if True keep all the "out-of-vocabulary" tokens, else ignore them

  """

  def __init__(self, is_alpha=False, is_digit=False, is_ascii=False,
               is_title=False):
    super(TYPEfilter, self).__init__()
    self.is_alpha = is_alpha
    self.is_digit = is_digit
    self.is_ascii = is_ascii
    self.is_title = is_title

  def filter(self, token, pos):
    flags = [
        self.is_alpha and token.isalpha(),
        self.is_digit and token.isdigit(),
        self.is_title and token.istitle(),
        self.is_ascii and all(ord(c) < 128 for c in token)
    ]
    if any(flags):
      return token
    return ''


class POSfilter(TokenFilter):
  """ we use universal tag-of-speech for filtering token
  NOUN : noun
  PRON : pronoun (I, me, he, she, herself, you, it, that, etc.)
  PROPN: proper noun (name of individual person, place, ... spelled with an initial capital letter, e.g. Jane, London)
  ADJ  : adjective
  VERB : verb
  ADV  : adverb

  ADP  : adposition (prepositional, postpositional, and circumpositional phrases)
  AUX  : auxiliary
  DET  : determiner
  INTJ : interjection
  NUM  : numeral
  PART : particle
  PUNCT: punctuation
  SCONJ: subordinating conjunction
  SYM  : symbol
  X    : other

  """

  def __init__(self, NOUN=True, PRON=False, PROPN=True,
               ADJ=True, VERB=False, ADV=True,
               ADP=False, AUX=False, DET=False, INTJ=False,
               NUM=False, PART=False, PUNCT=False,
               SCONJ=False, SYM=False, X=False):
    super(POSfilter, self).__init__()
    pos = []
    if NOUN: pos.append('NOUN')
    if PRON: pos.append('PRON')
    if PROPN: pos.append('PROPN')
    if ADJ: pos.append('ADJ')
    if ADP: pos.append('ADP')
    if ADV: pos.append('ADV')
    if AUX: pos.append('AUX')
    if DET: pos.append('DET')
    if INTJ: pos.append('INTJ')
    if NUM: pos.append('NUM')
    if PART: pos.append('PART')
    if PUNCT: pos.append('PUNCT')
    if SCONJ: pos.append('SCONJ')
    if SYM: pos.append('SYM')
    if VERB: pos.append('VERB')
    if X: pos.append('X')
    self.pos = pos

  def filter(self, token, pos):
    """
    Paramters
    ---------
    tokens: list
        pass

    Return
    ------
    list of accepted tokens
    """
    if pos is None or pos in self.pos:
      return token
    return ''


# ===========================================================================
# Preprocessing data
# ===========================================================================
# static variables for multiprocessing
def _preprocess_func(doc):
  preprocessors = globals()['__preprocessors']
  filters = globals()['__filters']
  charlevel = globals()['__charlevel']
  lang = globals()['__lang']
  lemma = globals()['__lemma']
  stopwords = globals()['__stopwords']

  doc_tokens = []
  # preprocessing document
  for p in preprocessors:
    doc = p(doc)
  # auto split if the doc haven't been splitted
  if isinstance(doc, string_types):
    doc = doc.split(' ')
  # ====== start processing ====== #
  for token in doc:
    if len(token) > 0:
      # ignore stopwords if requred
      if not stopwords and is_stopword(token, lang):
        continue
      # check if token is accepted
      if filters is not None:
        for f in filters:
          token = f(token, None)
      token = token.strip()
      # ignore token if it is removed
      if len(token) == 0: continue
      # normalize the token
      if lemma:
        pass
      # word-level dictionary
      if not charlevel:
        doc_tokens.append(token)
      # character-level dictionary
      else:
        for char in token:
          doc_tokens.append(char)
  return doc_tokens


class Tokenizer(object):

  """
  Parameters
  ----------
  preprocessors:
      pass
  filters:
      pass
  nb_words: int
      pass
  char_level: bool
      pass
  stopwords: bool
      if True, all stopwords are accepted, otherwise, remove all stopwords
      during preprocessing in `fit` function.
  order: 'word' or 'doc'
      if 'word', order the dictionary by word frequency
      if 'doc', order the dictionary by docs frequency (i.e. the number
      of documents that the word appears in)

  Note
  ----
  This module use `multiprocessing` to significantly speed up tokenizing
  process for big documents, but it might be slow on trivial dataset.

  """

  def __init__(self, nb_words=None,
               char_level=False,
               preprocessors=[TransPreprocessor(), CasePreprocessor(lower=True)],
               filters=None,
               stopwords=False,
               lemmatization=True,
               language='en',
               batch_size=2048,
               nb_processors=None,
               order='word',
               engine='odin',
               print_progress=True):
    # ====== internal states ====== #
    if engine not in ('spacy', 'odin'):
      raise ValueError('We only support 2 text processing engines: Spacy, or ODIN.')
    if order not in ('word', 'doc'):
      raise ValueError('The "order" argument must be "doc" or "word".')
    self.__engine = engine
    self.__order = order
    self.__longest_document = ['', 0]
    self.print_progress = print_progress
    # ====== dictionary info ====== #
    self._nb_words = nb_words
    self.nb_docs = 0
    self.char_level = char_level
    self.language = language

    self._word_counts = defaultdict(int)
    # number of docs the word appeared
    self._word_docs = defaultdict(int)
    # actual dictionary used for embedding
    self._word_dictionary = OrderedDict()
    self._word_dictionary_info = OrderedDict()

    self.stopwords = stopwords
    self.lemmatization = lemmatization

    self.batch_size = batch_size
    self.nb_processors = (int(nb_processors) if nb_processors is not None
                          else cpu_count())
    # ====== filter and preprocessor ====== #
    self.filters = filters if filters is None else as_tuple(filters)
    if preprocessors is None:
      preprocessors = []
    elif not isinstance(preprocessors, (tuple, list)):
      preprocessors = [preprocessors]
    self.preprocessors = preprocessors

  def _refresh_dictionary(self):
    # sort the dictionary
    word_counts = self._word_counts.items() if self.__order == 'word' \
        else self._word_docs.items()
    # sorted by both attribute for deterministic dictionary
    word_counts.sort(key=lambda x: (x[1], x[0]), reverse=True)
    # create the ordered dictionary
    word_dictionary = OrderedDict()
    word_dictionary_info = OrderedDict()
    word_dictionary[u''] = 0
    word_dictionary_info[0] = (0, 0) # word_index => (count, docs)
    # ====== add all tokens in order ====== #
    n = self.nb_words
    for i, (w, _) in enumerate(word_counts):
      if i + 1 >= n:
        break
      word_dictionary[w] = i + 1
      word_dictionary_info[i + 1] = (_, self._word_docs[w])
    self._word_dictionary = word_dictionary
    self._word_dictionary_info = word_dictionary_info
    return word_dictionary

  def _validate_texts(self, texts):
    """ Valiate the input to `fit` and `transform` """
    if not isinstance(texts, Iterable) and \
    not isinstance(texts, Iterator) and \
    not is_string(texts):
      raise ValueError('texts must be an iterator, generator or a string.')
    if is_string(texts):
      texts = (texts,)
    # convert to unicode
    texts = (t.decode('utf-8') for t in texts)
    return texts

  # ==================== properties ==================== #
  @property
  def nb_words(self):
    if self._nb_words is None or self._nb_words > len(self._word_counts):
      n = len(self._word_counts)
    else:
      n = int(self._nb_words)
    # always add the 0 words
    return n + 1

  @nb_words.setter
  def nb_words(self, value):
    if value != self._nb_words:
      self._nb_words = value
      self._refresh_dictionary()

  @property
  def longest_document(self):
    return self.__longest_document[0]

  @property
  def longest_document_length(self):
    return self.__longest_document[1]

  @property
  def summary(self):
    """ Return a dictionary with description -> statistics of
    the Tokenizer
    """
    return {
        '#Words': len(self._word_counts),
        '#Docs': self.nb_docs,
        'DictSize': len(self.dictionary),
        'longest_docs': self.longest_document_length,
        'Top': [i[0] for i in self.top_k(n=12)]
    }

  @property
  def dictionary(self):
    return self._word_dictionary

  def __len__(self):
    return len(self._word_counts)

  def top_k(self, n=12):
    top = []
    count = self._word_counts if self.__order == 'word' else self._word_docs
    for i, w in enumerate(self.dictionary.keys()):
      if i == 0: continue
      top.append((w, count[w]))
      if i >= n + 1: break
    return top

  # ==================== methods ==================== #
  def _preprocess_docs_spacy(self, texts, vocabulary, keep_order):
    def textit(texts):
      for t in texts:
        for p in self.preprocessors:
          t = p(t)
        if isinstance(t, (tuple, list)):
          t = ' '.join(t)
        yield t
    nlp = language(self.language)
    texts = textit(texts)
    for nb_docs, doc in enumerate(nlp.pipe(texts, n_threads=self.nb_processors,
                                           batch_size=self.batch_size)):
      doc_tokens = []
      for sent in doc.sents:
        for token in sent:
          # check if token is accepted
          token_POS = token.pos_
          # ignore stopwords if requred
          if not self.stopwords and is_stopword(token):
            continue
          # filtering
          if self.filters is not None:
            for f in self.filters:
              token = f(token.orth_, token_POS)
          # skip if token is removed by filter
          if len(token) == 0: continue
          # normalize the token
          if self.lemmatization and token_POS != 'PROPN': # lemmalization
            token = token.lemma_
          else:
            token = token.orth_
          token = token.strip()
          # word-level dictionary
          if not self.char_level:
            if vocabulary is None or token in vocabulary:
              doc_tokens.append(token)
          # character-level dictionary
          else:
            for char in token:
              if vocabulary is None or token in vocabulary:
                doc_tokens.append(char)
      yield nb_docs + 1, doc_tokens

  def _preprocess_docs_odin(self, texts, vocabulary, keep_order):
    # ====== main processing ====== #
    def initializer(filters, preprocessors,
                    lang, lemma, charlevel, stopwords):
      globals()['__preprocessors'] = preprocessors
      globals()['__filters'] = filters
      globals()['__lang'] = lang
      globals()['__lemma'] = lemma
      globals()['__charlevel'] = charlevel
      globals()['__stopwords'] = stopwords
    # add the index for ordering
    nb_docs = 0
    pool = Pool(processes=self.nb_processors, initializer=initializer,
                initargs=(self.filters, self.preprocessors, self.language,
                          self.lemmatization, self.char_level, self.stopwords))
    # return the tokenized documents as original order.
    if keep_order:
      it = pool.imap(func=_preprocess_func, iterable=texts,
                     chunksize=self.batch_size)
    # don't care about the order, often used for fitting
    else:
      it = pool.imap_unordered(func=_preprocess_func, iterable=texts,
                               chunksize=self.batch_size)
    # iterate over each return document
    for doc in it:
      nb_docs += 1
      if vocabulary is not None:
        doc = [token for token in doc if token in vocabulary]
      yield nb_docs, doc
    pool.close()
    pool.join()

  def fit(self, texts, vocabulary=None):
    """
    Parameters
    ----------
    texts: iterator of unicode
        iterator, generator or list (e.g. [u'a', u'b', ...])
        of unicode documents.
    """
    texts = self._validate_texts(texts)
    word_counts = self._word_counts
    word_docs = self._word_docs
    # ====== pick engine ====== #
    if self.__engine == 'spacy':
      processor = self._preprocess_docs_spacy
    elif self.__engine == 'odin':
      processor = self._preprocess_docs_odin
    # ====== start processing ====== #
    prog = Progbar(target=1208, name="Fitting tokenizer",
                   print_report=True, print_summary=True)
    start_time = timeit.default_timer()
    for nb_docs, doc in processor(texts, vocabulary, keep_order=False):
      total_docs_tokens = 0
      seen_words = {}
      # update words->count
      for token in doc:
        total_docs_tokens += 1
        word_counts[token] += 1
        # update words->doc
        if token not in seen_words:
          seen_words[token] = 1
          word_docs[token] += 1
      # save longest docs
      if total_docs_tokens > self.__longest_document[-1]:
        self.__longest_document = [doc, total_docs_tokens]
      # print progress
      prog['#Doc'] = nb_docs
      prog['#Tok'] = len(word_counts)
      prog.add(1)
      if prog.seen_so_far >= 0.8 * prog.target:
        prog.target = 1.2 * prog.target
    # ====== print summary of the process ====== #
    # if self.print_progress:
    #     prog.target = nb_docs; prog.update(nb_docs)
    processing_time = timeit.default_timer() - start_time
    print('Processed %d-docs, %d-tokens in %f second.' %
        (nb_docs, len(word_counts), processing_time))
    self.nb_docs += nb_docs
    # ====== sorting ====== #
    self._refresh_dictionary()
    return self

  # ==================== transforming odin ==================== #
  def transform(self, texts, mode='seq', dtype='int32',
                padding='pre', truncating='pre', value=0.,
                end_document=None, maxlen=None,
                token_not_found='ignore'):
    """
    Parameters
    ----------
    texts: iterator of unicode
        iterator, generator or list (e.g. [u'a', u'b', ...])
        of unicode documents.
    mode: 'binary', 'tfidf', 'count', 'freq', 'seq'
        'binary', abc
        'tfidf', abc
        'count', abc
        'freq', abc
        'seq', abc
    token_not_found: 'ignore', 'raise', a token string, an integer
        pass
    """
    # ====== check arguments ====== #
    texts = self._validate_texts(texts)
    # ====== check mode ====== #
    mode = str(mode)
    if mode not in ('seq', 'binary', 'count', 'freq', 'tfidf'):
      raise ValueError('The "mode" argument must be: "seq", "binary", '
                       '"count", "freq", or "tfidf".')
    # ====== check token_not_found ====== #
    if not is_number(token_not_found) and \
    not is_string(token_not_found) and \
    token_not_found not in ('ignore', 'raise'):
      raise ValueError('token_not_found can be: "ignore", "raise"'
                       ', an integer of token index, or a string '
                       'represented a token.')
    if token_not_found not in ('ignore', 'raise'):
      token_not_found = int(self.dictionary[token_not_found])
    elif is_number(token_not_found):
      token_not_found = int(token_not_found)
    # ====== pick engine ====== #
    if self.__engine == 'spacy':
      processor = self._preprocess_docs_spacy
    elif self.__engine == 'odin':
      processor = self._preprocess_docs_odin
    # ====== Initialize variables ====== #
    dictionary = self.dictionary
    results = []
    # ====== preprocess arguments ====== #
    if isinstance(end_document, str):
      end_document = dictionary.index(end_document)
    elif is_number(end_document):
      end_document = int(end_document)
    # ====== processing ====== #
    if hasattr(texts, '__len__'):
      target_len = len(texts)
      auto_adjust_len = False
    else:
      target_len = 1208
      auto_adjust_len = True
    prog = Progbar(target=target_len, name="Tokenize Transform",
                   print_report=True, print_summary=True)
    for nb_docs, doc in processor(texts, vocabulary=None, keep_order=True):
      # found the word in dictionary
      vec = []
      for x in doc:
        idx = dictionary.get(x, -1)
        if idx >= 0: vec.append(idx)
        # not found the token in dictionary
        elif token_not_found == 'ignore':
          continue
        elif token_not_found == 'raise':
          raise RuntimeError('Cannot find token: "%s" in dictionary' % x)
        elif isinstance(token_not_found, int):
          vec.append(token_not_found)
      # append ending document token
      if end_document is not None:
        vec.append(end_document)
      # add the final results
      results.append(vec)
      # print progress
      if self.print_progress:
        prog['#Docs'] = nb_docs
        prog.add(1)
        if auto_adjust_len and prog.seen_so_far >= 0.8 * prog.target:
          prog.target = 1.2 * prog.target
    # end the process
    # if self.print_progress and auto_adjust_len:
    #     prog.target = nb_docs; prog.update(nb_docs)
    # ====== pad the sequence ====== #
    # just transform into sequence of tokens
    if mode == 'seq':
      maxlen = self.longest_document_length if maxlen is None \
          else int(maxlen)
      results = pad_sequences(results, maxlen=maxlen, dtype=dtype,
                              padding=padding, truncating=truncating,
                              value=value)
    # transform into one-hot matrix
    else:
      X = np.zeros(shape=(len(results), self.nb_words))
      for i, seq in enumerate(results):
        if mode == 'binary':
          X[i, seq] = 1
        elif mode == 'freq':
          length = len(seq)
          count = freqcount(seq)
          for tok, n in count.items():
            X[i, tok] = n / float(length)
        elif mode == 'count':
          count = freqcount(seq)
          for tok, n in count.items():
            X[i, tok] = n
        elif mode == 'tfidf':
          count = freqcount(seq)
          for tok, n in count.items():
            tf = 1 + np.log(n)
            docs_freq = self._word_dictionary_info.get(tok, (0, 0))[-1]
            idf = np.log(1 + self.nb_docs / (1 + docs_freq))
            X[i, tok] = tf * idf
      results = X
    return results

  def embed(self, vocabulary, dtype='float32',
            token_not_found='ignore'):
    """Any word not found in the vocabulary will be set to all-zeros"""
    # ====== check vocab ======= #
    if not isinstance(vocabulary, Mapping):
      raise ValueError('"vocabulary" must be any instance of dict.')
    # ====== check token_not_found ====== #
    if not is_number(token_not_found) and \
    not is_string(token_not_found) and \
    token_not_found not in ('ignore', 'raise'):
      raise ValueError('token_not_found can be: "ignore", "raise"'
                       ', an integer of token index, or a string '
                       'represented a token.')
    if token_not_found not in ('ignore', 'raise'):
      token_not_found = int(self.dictionary[token_not_found])
    elif is_number(token_not_found):
      token_not_found = int(token_not_found)
    # ====== create embedding matrix ====== #
    ndim = len(next(vocabulary.values()))
    matrix = np.zeros(shape=(len(self.dictionary), ndim), dtype=dtype)
    for word, idx in self.dictionary.items():
      if len(word) == 0: continue
      if word in vocabulary:
        matrix[idx, :] = vocabulary[word]
      elif token_not_found == 'raise':
        raise Exception('Cannot find token "%s" in the vocabulary.' % word)
      elif isinstance(token_not_found, int):
        matrix[idx, :] == matrix[token_not_found, :]
    return matrix



from __future__ import print_function, absolute_import, division

from odin.nnet.base import *
from odin.nnet.math_utils import *
from odin.nnet.conv import *
from odin.nnet.noise import *
from odin.nnet.shape import *
from odin.nnet.sampling import *
from odin.nnet.normalization import *
from odin.nnet.embedding import *
from odin.nnet.helper import *
from odin.nnet.rnn import *
from odin.nnet.time_delayed import *

from odin.nnet import models

from odin.utils import uuid, bin2folder, folder2bin

# ===========================================================================
# Helper method for serialize NNOp
# ===========================================================================
def serialize(nnops, path=None, save_variables=True, variables=[],
              binary_output=False, override=False):
  """ Serialize NNOp or list of NNOp and all necessary variables
  to a folder.

  Parameters
  ----------
  nnops: NNOp, Object, or list; tuple of NNOp and Object
  path: str
      path to a folder
  save_variables: bool
      if True, save all variables related to all given NNOps
  variables: list of tensorflow Variables
      additional list of variables to be saved with this model
  binary_output: bool (default: False)
      if `False` (by default), original way tensorflow serialize
      all variables, save all variables and nnop info to separated
      files within a folder `path`
      if `True`, convert all files in the folder to binary
      and save to a dictionary with its relative path, if
      `path` is not None, use pickle to save all binary data
      to a file
  override: bool
      if True, remove existed folder to override everything.

  Return
  ------
  path: str
      path to the folder that store NNOps and variables
  """
  # ====== check output_mode ====== #
  if path is None and not binary_output:
    raise ValueError('`path` cannot be None if `binary_output=False`')
  is_path_given = False if path is None else True
  if path is None:
    path = '/tmp/tmp' # default path
  path_folder = path + uuid(length=25) if binary_output else path
  # ====== getting save data and variables ====== #
  vars = []
  if save_variables:
    for op in as_tuple(nnops):
      if hasattr(op, 'variables'):
        for v in as_tuple(op.variables):
          if K.is_variable(v):
            vars.append(v)
  vars = list(set(vars + as_list(variables)))
  # ====== checking path ====== #
  # It is important to remove the `path_folder` AFTER getting all
  # the variables, since this can remove the path to restored
  # variables required in `op.variables`
  if os.path.exists(path_folder):
    if os.path.isfile(path_folder):
      raise ValueError("path: '%s' is NOT a folder." % path_folder)
    elif override:
      shutil.rmtree(path_folder)
      os.mkdir(path_folder)
  else:
    os.mkdir(path_folder)
  nnops_path = os.path.join(path_folder, 'nnops.ai')
  vars_path = os.path.join(path_folder, 'variables')
  # save NNOps
  with open(nnops_path, 'wb') as f:
    cPickle.dump(nnops, f, protocol=cPickle.HIGHEST_PROTOCOL)
  # save Variables
  if len(vars) > 0:
    K.save_variables(vars, vars_path)
  # ====== convert folder to file or binary ====== #
  if binary_output:
    data = folder2bin(path_folder)
    # only return binary data
    if not is_path_given:
      shutil.rmtree(path_folder)
      return data
    # given path, save binary to path
    # check if override
    if os.path.exists(path):
      if override:
        if os.path.isfile(path):
          os.remove(path)
        else:
          shutil.rmtree(path)
      else:
        raise RuntimeError("File at path: %s exists, cannot override." % path)
    # write file
    with open(path, 'wb') as f:
      cPickle.dump(data, f, protocol=cPickle.HIGHEST_PROTOCOL)
    shutil.rmtree(path_folder)
  return path


def deserialize(path_or_data, force_restore_vars=True):
  """
  Parameters
  ----------
  path_or_data : {string, dict}
      if a path is given (i.e. string types), load dumped model from
      given folder
      if a dictionary is given, load binary data directly

  force_restore_vars : bool (default=True)
      if `False`, this is special tricks, the unpickled NNOp stay useless
      until its variables are restored,
      but if we restore the variables right away, it create a
      session and prevent any possibility of running
      tensorflow with multiprocessing
      => store the `_restore_vars_path` in NNOp for later,
      and restore the variable when the NNOp is actually in used.

  Note
  ----
  if `force_restore_vars = False`, this create 1 flaw,
  if the nested NNOp is called before the unpickled NNOp
  restore its variables, the nested Ops cannot acquire its
  variables.

  """
  data = None
  path_folder = '/tmp/tmp_%s' % uuid(12)
  delete_after = True
  # ====== check path ====== #
  if is_string(path_or_data):
    # path to a file
    if os.path.isfile(path_or_data):
      with open(path_or_data, 'rb') as f:
        data = cPickle.load(f)
    # path to a folder
    elif os.path.isdir(path_or_data):
      path_folder = path_or_data
      delete_after = False
    else: # pickle string
      data = cPickle.loads(path_or_data)
  # given data
  elif isinstance(path_or_data, dict):
    data = path_or_data
  # ====== check data ====== #
  if data is not None:
    bin2folder(data, path=path_folder)
  path = path_folder
  # ====== read normally from folder ====== #
  nnops_path = os.path.join(path, 'nnops.ai')
  vars_path = os.path.join(path, 'variables')
  # ====== load the NNOps ====== #
  if not os.path.exists(nnops_path):
    raise ValueError("Cannot file path to serialized NNOps at: %s" % nnops_path)
  with open(nnops_path, 'rb') as f:
    nnops = cPickle.load(f)
  # ====== load the Variables ====== #
  if os.path.exists(vars_path + '.index'):
    if force_restore_vars:
      K.restore_variables(vars_path)
      # delete cached folder
      if delete_after:
        shutil.rmtree(path)
    else:
      nnops._set_restore_info(vars_path, delete_after)
  return nnops



"""
MIT License
===========

Copyright (c) 2012 TrungNT (email: [name]@imito.ai)

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import print_function, division, absolute_import

import os
from six.moves import zip, zip_longest, range, cPickle
from collections import Mapping

import numpy as np

from odin.utils import (segment_list, one_hot, is_string, Progbar, batching,
                        as_tuple, ctext, is_number, is_primitives,
                        defaultdictkey)
from odin.utils.mpi import MPI, async
from odin.fuel.data import Data, as_data
from odin.fuel.recipe_base import RecipeList

# ===========================================================================
# Helper for grouping
# ===========================================================================
def _to_numpy_array(self, x):
  if not is_string(x[0]) and len(set(i.shape[1:] for i in x)) == 1:
    return np.concatenate(x, axis=0)
  return np.array(x)


def _batch_grouping(batch, batch_size, rng, batch_filter):
  """ batch: contains
      [
          (name, [list of data]),
          (name, [list of data]),
          (name, [list of data]),
          ...
      ]

  Note
  ----
  We assume the shape[0] (or length) of all "data" and "others" are
  the same
  """
  if len(batch) == 0:
    yield None
  else:
    # create batch of indices for each file (indices is the start
    # index of each batch)
    indices = [list(range(0, X[0].shape[0], batch_size))
               for name, X in batch]
    # shuffle if possible
    if rng is not None:
      [rng.shuffle(i) for i in indices]
    # ====== create batch of data ====== #
    for idx in zip_longest(*indices):
      ret = []
      for start, (name, X) in zip(idx, batch):
        # skip if the one data that is not enough
        if start is None: continue
        # pick data from each given input
        end = start + batch_size
        _ = [x[start:end] for x in X]
        ret.append(_)
      ret = [np.concatenate(x, axis=0) for x in zip(*ret)]
      # shuffle 1 more time
      N = list(set([r.shape[0] for r in ret]))
      if len(N) > 1:
        raise ValueError("The shape[0] of Data is different, found "
                         "%d different length: %s" % (len(N), str(N)))
      N = N[0]
      if rng is not None:
        permutation = rng.permutation(N)
        ret = [r[permutation] for r in ret]
      # return the batches
      for start in range(0, N, batch_size):
        end = start + batch_size
        _ = batch_filter([x[start:end] for x in ret])
        # always return tuple or list
        if _ is not None:
          yield _ if isinstance(_, (tuple, list)) else (ret,)

def _file_grouping(batch, batch_size, rng, batch_filter):
  """ Return: [(name, index, data1, data2, ...), ...]
      NOTE: each element in batch is one file
  """
  # ====== shuffle the file ====== #
  if rng is not None:
    rng.shuffle(batch)
  # ====== return batched files with index for ordering ====== #
  for name, X in batch:
    n = X[0].shape[0]
    ret = list(X)
    for i, (start, end) in enumerate(batching(n=n, batch_size=batch_size)):
      r = [name, i] + [j[start:end] for j in ret]
      yield tuple(batch_filter(r))

# ===========================================================================
# IndexedData
# ===========================================================================
_apply_approx = lambda n, x: int(round(n * x)) if x < 1. + 1e-12 else int(x)
_indices_dtype = [('name', 'object'), ('start', 'i4'), ('end', 'i4')]

def _preprocessing_indices(indices):
  """ Three different kind of indices:
  * file: load from csv file with ' ' (i.e. space as separator)
  * array: numpy array
  * nosql: instance of NoSQL
  """
  # store information for reloading the indices
  indices_info = None
  # indices always sorted in [(name, start, end), ...]
  if isinstance(indices, str) and os.path.isfile(indices):
    indices = {str(name): (start, end) for name, start, end in
               np.genfromtxt(indices, dtype=_indices_dtype, delimiter=' ')}
    indices_info = ('file', str(indices))
  # list or tuple form: (name, (start, end)) or dictionary
  else:
    if isinstance(indices, (tuple, list, np.ndarray)):
      if len(indices[0]) == 2:
        indices = {name: (int(start), int(end))
                   for name, (start, end) in indices}
      else:
        indices = {name: (int(start), int(end))
                   for name, start, end in indices}
    # dictionary: name -> (start, end) or (name, start, end)
    elif isinstance(indices, Mapping):
      pass
    else:
      raise ValueError('Unsupport `indices` type: "%s".' % type(indices))
    indices_info = ('mapping', indices)
  return indices, indices_info

class IndexedData(Data):
  """

  Parameters
  ----------
  data : {Data, list of Data}
    list of Data will be manipulated by this descriptor
    NOTE: all Data must have the same length
  indices : Mapping
    mapping from `name`->(start, end)

  """

  def __init__(self, data, indices):
    super(IndexedData, self).__init__(data=data, read_only=True)
    # ====== states variables ====== #
    self._length = None
    # if True return name during __iter__
    self._return_name = False
    # ====== load indices ====== #
    self._indices_loader = async(_preprocessing_indices)(indices)
    self._indices_info = None
    self._indices = None # dictionary: name -> (start, end)
    # ====== Load data ====== #
    # check all data have the same shape[0]
    length = len(self.data[0])
    if any(d.shape[0] != length for d in self.data):
      raise ValueError('All Data must have the same length '
                       '(i.e. shape[0]), the given data have '
                       'shape: %s' % str([d.shape for d in data]))

  # ==================== Properties ==================== #
  @property
  def indices_info(self):
    if self._indices_info is None:
      self._indices, self._indices_info = self._indices_loader.get()
    return self._indices_info

  @property
  def indices(self):
    if self._indices is None:
      self._indices, self._indices_info = self._indices_loader.get()
    return self._indices

  @property
  def nb_files(self):
    return len(self._indices)

  # ==================== pickling ==================== #
  @property
  def data_info(self):
    return (self._indices_info, self._data,
            self._length, self._return_name)

  def _restore_data(self, info):
    (self._indices_info, self._data,
        self._length, self._return_name) = info
    # deserialize indices
    ids_type, info = self._indices_info
    if ids_type == 'mapping':
      self._indices = info
    elif ids_type == 'file':
      self._indices = np.genfromtxt(info, dtype=_indices_dtype,
                                    delimiter=' ')

  # ==================== override from Data ==================== #
  @property
  def shape(self):
    """ This is just an "UPPER" estimation, some data points might be lost
    during preprocessing each indices by recipes.
    """
    if self._length is None:
      self._length = sum((end - start)
                      for name, (start, end) in self.indices.items())
    ret_shape = [(self._length,) + dat.shape[1:]
                 for dat in self.data]
    return tuple(ret_shape) if self.is_data_list else ret_shape[0]

  def __str__(self):
    name = ctext('IndexedData', 'cyan')
    s = '<%s: Indices(type:"%s" length:%d)>\n' % \
        (name, self.indices_info[0], len(self.indices))
    for dat in self.data:
      s += '   (%s)%s: %s %s\n' % \
          (dat.__class__.__name__,
              ctext(str(dat.data_info), 'yellow'),
              dat.shape,
              str(dat.dtype))
    return s[:-1]

  # ==================== Strings ==================== #
  def keys(self):
    return self.indices.keys()

  def __getitem__(self, key):
    if is_string(key):
      key = slice(*self.indices[key])
    return super(IndexedData, self).__getitem__(key)

# ===========================================================================
# Multiprocessing Feeder
# ===========================================================================
def _dummy_batch_filter(x):
  return x


class Feeder(Data):
  """ multiprocessing Feeder to 1 comsumer
  Process1    Process2 ...    Process3
      |          |     |          |
       ------- Map Function ------
      |          |     |          |
       ----- Reduce Function -----
       \            |            /
         ---- Return batches ---

  This feeder return a non-deterministic order of data, hence,
  cannot be reproducible

  map_function: (name, x, transcription)
  reduce_function: (list of objects returned from map_function)

  Parameters
  ----------
  indices: path(csv file), list, ndarray, dict
      indices represent following information: [name, start_id, end_id]
      if indices is dictionary, it must in the form: {name: (start, end)}
  dtype: string or numpy.dtype
      all data return from this feeder are converted to given dtype
      if None, original dtype is kept
  batch_filter: call-able
      must be a function has take a list of np.ndarray as first arguments
      ([X]) or ([X, y]), you can return None to ignore given batch, return the
      data for accepting the batch
  batch_mode: 'batch' or 'file' (string type)
      'batch' mode return shuffling and return everything in small batches
      'file' mode return [(file_name, order_index_from_0, data...), ...]
  ncpu: int
      number of CPU used for multiprocessing
  buffer_size: int
      A process will perform processing on a group of `buffer_size` number of
      data points, then, a list of results are returned to the main process.
      The higher this number the more powerful batch shuffling.
  hwm: int
      "high water mark" for SEND socket, is a hard limit on the
      maximum number of outstanding messages ØMQ shall queue
      in memory for any single peer that the specified socket
      is communicating with.
  mpi_backend: {'pyzmq', 'python'}
      the A.P.I for message passing between process, sometimes
      python Queue is faster than pyZMQ, but if pyZMQ is faster
      it could up to 35%.

  Example
  -------
  >>> ds = F.Dataset(os.path.join(temppath, 'ds'), read_only=True)
  >>> feeder = F.Feeder(ds['X'], indices=ds['indices.csv'],
  >>>                   ncpu=2, buffer_size=2, hwm=12)
  >>> feeder.set_recipes([
  >>>     F.recipes.TransLoader(ds['transcription.dict'], dtype='int32'),
  >>>     F.recipes.CreateBatch()
  >>> ])
  >>> for i, j in feeder.set_batch(12, seed=1208251813, shuffle_level=2):
  >>>     for x, y in zip(i, j):
  >>>         print(transcription_test[str(x.tolist())] == y)

  Note
  ----
  set(ncpu=1) if you want a reproducible results
   - Memory transferring in Queue is always the bottleneck of multiprocessing
  3 supporting mode for shuffling:
   - shuffle_level=0: only shuffling the indices
   - shuffle_level=1: shuffle the buffered batch (e.g. 12 files in the indices)
   - shuffle_level=2: shuffle each returned batch
  * you must balance 2 number: buffer_size and hwm, so the
  amount of data cached by all processed does not excess the RAM

  """

  def __init__(self, data_desc, dtype=None,
               batch_filter=None, batch_mode='batch',
               ncpu=1, buffer_size=8, hwm=86,
               mpi_backend='python'):
    super(Feeder, self).__init__(data=as_tuple(data_desc, t=IndexedData),
                                 read_only=True)
    # find intersection of all indices in IndexedData
    self._indices_keys = async(
        lambda: np.array(
            list(set.intersection(*[set(dat.indices.keys())
                                    for dat in self._data])),
            dtype=str)
    )()
    # ====== desire dtype ====== #
    nb_data = sum(len(dat._data) for dat in self._data)
    self._output_dtype = as_tuple(dtype, N=nb_data)
    # ====== Set default recipes ====== #
    self._recipes = RecipeList()
    self._recipes.set_feeder_info(nb_desc=len(self._data))
    self.set_multiprocessing(ncpu, buffer_size, hwm, mpi_backend)
    # ====== cache shape information ====== #
    # store first dimension
    self._cache_shape = None
    # if the recipes changed the shape need to be recalculated
    self._recipes_changed = False
    # ====== Iteration information ====== #
    self._running_iter = []
    # ====== batch mode ====== #
    if batch_filter is None:
      batch_filter = _dummy_batch_filter
    elif not hasattr(batch_filter, '__call__'):
      raise ValueError('batch_filter must be a function has 1 or 2 '
                       'parameters (X) or (X, y).')
    # check if batch_filter Picklable
    try:
      cPickle.dumps(batch_filter, protocol=2)
    except Exception:
      raise ValueError("`batch_filter` must be pickle-able, which must be "
                       "top-level function.")

    self._batch_filter = batch_filter
    # check batch_mode
    batch_mode = str(batch_mode).lower()
    if batch_mode not in ("batch", 'file'):
      raise ValueError("Only support `batch_mode`: 'file'; 'batch', but "
                       "given value: '%s'" % batch_mode)
    self._batch_mode = batch_mode

  # ==================== pickling ==================== #
  @property
  def data_info(self):
    return (self._data, self.indices_keys, self._recipes,
            self._output_dtype, self._cache_shape,
            self._batch_mode, self._batch_filter,
            self.ncpu, self.buffer_size, self.hwm)

  def _restore_data(self, info):
    (self._data, self._indices_keys, self._recipes,
     self._output_dtype, self._cache_shape,
     self._batch_mode, self._batch_filter,
     self.ncpu, self.buffer_size, self.hwm) = info
    # ====== basic attributes ====== #
    self._recipes_changed = False
    self._running_iter = []

  # ==================== multiprocessing ==================== #
  def set_multiprocessing(self, ncpu, buffer_size=None, hwm=None,
                          mpi_backend=None):
    self.ncpu = ncpu
    if buffer_size is not None:
      self.buffer_size = int(buffer_size)
    if hwm is not None:
      self.hwm = int(hwm)
    if mpi_backend is not None:
      self.mpi_backend = str(mpi_backend)
    return self

  def set_batch(self, batch_size=None, batch_filter=None, batch_mode=None,
                seed=-1, start=None, end=None, shuffle_level=None):
    # ====== check batch_filter ====== #
    if batch_filter is not None:
      if not hasattr(batch_filter, '__call__'):
        raise ValueError(
            'batch_filter must be a function has 1 arguments (X)')
      self._batch_filter = batch_filter
    # ====== check batch_mode ====== #
    if batch_mode is not None:
      batch_mode = str(batch_mode).lower()
      if batch_mode not in ("batch", 'file'):
        raise ValueError("Only support `batch_mode`: 'file'; 'batch', but "
                         "given value: '%s'" % batch_mode)
      self._batch_mode = batch_mode
    return super(Feeder, self).set_batch(batch_size=batch_size, seed=seed,
                                         start=start, end=end,
                                         shuffle_level=shuffle_level)

  def set_recipes(self, *recipes):
    self._recipes_changed = True
    self._recipes.set_recipes(recipes)
    self.shape # re-calculate cached shape
    return self

  # ==================== override from Data ==================== #
  @property
  def nb_files(self):
    return len(self.indices_keys)

  @property
  def indices_keys(self):
    if not isinstance(self._indices_keys, np.ndarray):
      self._indices_keys = self._indices_keys.get()
    return self._indices_keys

  @property
  def dtype(self):
    """ This is only return the desire dtype for input
    Data, not the ones outputed by Feeder. """
    all_dtypes = []
    for dat in self._data:
      for d in dat._data:
        all_dtypes.append(d.dtype)
    return tuple([j if i is None else i
                  for i, j in zip(self._output_dtype, all_dtypes)])

  @property
  def shape(self):
    """ This is just an "UPPER" estimation, some data points might be lost
    during preprocessing each indices by recipes.
    """
    # ====== first time calculate the shape ====== #
    if self._cache_shape is None or self._recipes_changed:
      # for each Descriptor, create list of pairs: (name, length)
      shapes_indices = []
      for dat in self._data:
        indices = []
        length = 0
        for name in self.indices_keys:
          start, end = dat.indices[name]
          lng = end - start
          length += lng
          indices.append((name, lng))
        # modify shapes by estimted length from indices
        shapes = (dat.shape,) if is_number(dat.shape[0]) \
            else dat.shape
        # NOTE: the indices is copy for each shape (i.e. data),
        # hence, it will create some overhead in shape_transform
        for shp in [(length,) + shp[1:] for shp in shapes]:
          shapes_indices.append((shp, list(indices)))
      # Recipes shape_transform
      shapes = tuple([
          shp for shp, ids in self._recipes.shape_transform(shapes_indices)
      ])
      del shapes_indices
      self._cache_shape = tuple(shapes)
      self._recipes_changed = False
    # ====== get the cached shape ====== #
    if any(s[0] == 0 for s in self._cache_shape):
      raise RuntimeError("Feeder has `length=0` change the recipes to retain "
                         "minimum of `length>=1`, shape: %s" % str(self._cache_shape))
    return self._cache_shape

  def __str__(self):
    padding = '   '
    s = '<%s: #keys:%d #iter:%d #CPU:%s #Buffer:%d #HWM:%d mode:"%s">\n' % \
        (ctext('Feeder', 'cyan'), len(self.indices_keys),
            len(self._running_iter), self.ncpu, self.buffer_size,
            self.hwm, self._batch_mode)
    # ====== Shape and dtype ====== #
    shape = self.shape # this is always list of shape
    s += padding + ctext("Shape: ", 'magenta') + \
        ', '.join((str(s) for s in shape)) + '\n'
    s += padding + ctext("Dtype: ", 'magenta') + \
        ', '.join((str(dt) for dt in self.dtype)) + '\n'
    # ====== print recipes ====== #
    s += padding + ctext('Recipes:', 'magenta') + '\n'
    for recipe in self._recipes:
      s += '\n'.join(['\t' + i for i in str(recipe).split('\n')])
      s += '\n'
    # ====== print data descriptor ====== #
    s += padding + ctext('Descriptor:', 'magenta') + '\n'
    for desc in self._data:
      s += '\n'.join(['\t' + i for i in str(desc).split('\n')])
      s += '\n'
    return s[:-1]

  # ==================== Strings ==================== #
  def __iter__(self):
    # ====== get start and end for indices ====== #
    start = _apply_approx(self.nb_files, self._start)
    end = _apply_approx(self.nb_files, self._end)
    all_keys = self.indices_keys[start:end]
    # ====== shuffle the indices ====== #
    rng = None
    shuffle_level = self._shuffle_level
    if self._seed is not None:
      rng = np.random.RandomState(self._seed)
      all_keys = all_keys[rng.permutation(self.nb_files)]
      if shuffle_level < 1:
        rng = None
      # reset the seed
      self._seed = None
    batch_size = self._batch_size
    batch_filter = self._batch_filter
    process_func = self._recipes.process
    # ====== prepare data, indices and dtype ====== #
    data_indices_dtype = []
    i = 0
    for dat in self._data:
      for d in dat._data:
        data_indices_dtype.append(
            (d, dat.indices, self._output_dtype[i]))
        i += 1

    # ====== create wrapped functions ====== #
    def map_func(jobs):
      if self.buffer_size == 1:
        jobs = [jobs]
      # calculating batch results
      batch = []
      for name in jobs:
        X = []
        for dat, ids, dtype in data_indices_dtype:
          start, end = ids[name]
          # data can be list of Data, or just 1 Data
          dat = dat[start:end]
          if dat.dtype != dtype:
            dat = dat.astype(dtype)
          X.append(dat)
        X = process_func(name, X)
        # ignore None returned result
        if X is not None:
          batch.append(X)
      # choose grouping function
      if self._batch_mode == 'batch':
        X = _batch_grouping(batch, batch_size, rng, batch_filter)
      elif self._batch_mode == 'file':
        X = _file_grouping(batch, batch_size, rng, batch_filter)
      return X

    # ====== track and return ====== #
    it = MPI(jobs=all_keys, func=map_func, ncpu=self.ncpu,
             batch=self.buffer_size, hwm=self.hwm,
             backend=self.mpi_backend)
    self._running_iter.append(it)
    return iter(it)

  def save_cache(self, path, name=None, dtype=None, batch_size=1024):
    """ Save all preprocessed data to a Dataset

    Parameters
    ----------
    path: string
        path to a folder
    name: None, or list of string
        specific name for each returned `numpy.ndarray` during iteration
    dtype: None, or list of dtype, or single dtype
        specific dtype for all or each of returned `numpy.ndarray`
        during iteration
    batch_size: int
        amount of samples for each batch (higher the faster iteration)

    Note
    ----
    Only returned `numpy.ndarray` are saved
    """
    from odin.fuel.dataset import Dataset
    if not is_string(path):
      raise ValueError("`path` must be string path to a folder.")
    if os.path.exists(path) and os.path.isfile(path):
      raise ValueError("`path` is a file, required a folder for "
                       "saving all cache data.")
    # ====== start caching ====== #
    prog = Progbar(target=len(self),
                   name='Saving cache of preprocessed data',
                   print_report=True, print_summary=True)
    ds = Dataset(path, override=True)
    with self.set_batch_context(batch_size=int(batch_size), seed=None,
                                start=0, end=-1, shuffle_level=0):
      for X in self:
        if not isinstance(X, (tuple, list)):
          X = (X,)
        n = 0
        i = 0
        # saving preprocessed data
        for x in X:
          if isinstance(x, np.ndarray):
            # checking name
            if name is None:
              x_name = 'X%d' % i
            else:
              x_name = name[i]
            # checking dtype
            if isinstance(dtype, (tuple, list)):
              x = x.astype(dtype[i])
            elif dtype is not None:
              x = x.astype(dtype)
            # saving to the dataset
            if x_name in ds:
              ds[x_name].append(x)
            else:
              ds[(x_name, 'memmap')] = x
            # update samples count, and data count
            n = x.shape[0]
            i += 1
        # print progress
        prog.add(n)
    # ====== flush and close everything ====== #
    ds.flush()
    ds.close()
    with open(os.path.join(path, 'README'), 'wb') as f:
      f.write(str(self))
    # end
    # ====== check one more time ====== #
    ds = Dataset(path, read_only=True)
    print(ds)
    print(ctext("Dataset size:", 'cyan'), ds.size, '(MB)')
    ds.close()
    return self

  def stop_all(self):
    """ Call this method to stop all processes in case you
    spamming to many iteration
    """
    for i in self._running_iter:
      i.terminate()
    self._running_iter = []

  def __del__(self):
    self.stop_all()
